/*
 * (C) Copyright IBM Corp. 2003
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   Konrad Rzeszutek <konradr@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.15.2.1 2003/02/27 05:59:41 rstory Exp $
 *
 * $Id: saHpiSensorTable.c,v 1.26 2004/04/27 15:55:56 konradrzeszutek Exp $
 *
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>
#include <netinet/in.h>
#include <saHpiSensorTable.h>

#include <saHpiSensorReadingCurrentTable.h>
#include <saHpiSensorReadingMinTable.h>
#include <saHpiSensorReadingMaxTable.h>
#include <saHpiSensorReadingNominalTable.h>
#include <saHpiSensorReadingNormalMinTable.h>
#include <saHpiSensorReadingNormalMaxTable.h>

#include <saHpiSensorThdLowCriticalTable.h>
#include <saHpiSensorThdLowMajorTable.h>
#include <saHpiSensorThdLowMinorTable.h>

#include <saHpiSensorThdUpCriticalTable.h>
#include <saHpiSensorThdUpMajorTable.h>
#include <saHpiSensorThdUpMinorTable.h>

#include <saHpiSensorThdPosHysteresisTable.h>
#include <saHpiSensorThdNegHysteresisTable.h>

#include <hpiSubagent.h>
#include <SaHpi.h>

static netsnmp_handler_registration *my_handler = NULL;
static netsnmp_table_array_callbacks cb;


static oid saHpiSensorTable_oid[] = { saHpiSensorTable_TABLE_OID };
static size_t saHpiSensorTable_oid_len = OID_LENGTH (saHpiSensorTable_oid);

static oid saHpiSensorCount_oid[] = { hpiResources_OID, 5, 0 };


static u_long sensor_count = 0;
static sensor_range_flags range_flags[] = {
  {SAHPI_SRF_MIN, "MIN, "},
  {SAHPI_SRF_MAX, "MAX, "},
  {SAHPI_SRF_NORMAL_MIN, "NORMAL_MIN, "},
  {SAHPI_SRF_NORMAL_MAX, "NORMAL_MAX, "},
  {SAHPI_SRF_NOMINAL, "NOMINAL, "}
};

#define RANGE_FLAGS_LEN 5


static int
saHpiSensorTable_modify_context (SaHpiEntryIdT rdr_id,
				 SaHpiSensorRecT * entry,
				 SaHpiSensorEvtEnablesT * enables,
				 SaHpiRptEntryT * rpt_entry,
				 oid *, size_t,
				 saHpiSensorTable_context * ctx);




int
populate_sensor (SaHpiEntryIdT rdr_id,
		 SaHpiSensorRecT * sensor,
		 SaHpiRptEntryT * rpt_entry,
		 oid * rdr_entry_oid, size_t rdr_entry_oid_len,
		 oid * sensor_oid, size_t * sensor_oid_len)
{

  int rc = AGENT_ERR_NOERROR;

  oid index_oid[SENSOR_INDEX_NR];
  oid column[2];

  netsnmp_index sensor_index;
#ifdef BUG_873961
  int i = 0;
  netsnmp_void_array *array;
  saHpiSensorTable_context *ctx;
#endif
  saHpiSensorTable_context *sensor_context;
  SaHpiSensorThresholdsT sensor_threshold;
  SaHpiSessionIdT session_id;
  SaHpiSensorEvtEnablesT enables;
  SaHpiSensorReadingT current_reading;

  DEBUGMSGTL ((AGENT, "\n\t--- populate_sensor: Entry.\n"));

  if (sensor)
    {
      sensor_index.len = SENSOR_INDEX_NR;
      // Look at the MIB to find out what the indexs are
      index_oid[0] = rpt_entry->DomainId;
      index_oid[1] = rpt_entry->ResourceId;
      index_oid[2] = sensor->Num;

      sensor_index.oids = (oid *) & index_oid;
      // We are re-populating. Check for existing entries
      sensor_context = NULL;
      sensor_context = CONTAINER_FIND (cb.container, &sensor_index);
      // If we don't find it - we create it.
#ifdef  BUG_873961
      if (!sensor_context)
	{
	  // Bug # 873961. We use the 'rdr_id' to check to see if
	  // it is the context. To do so, we have to search for it first.
	  sensor_index.len = 1;
	  // re-using the index_oid.
	  array = CONTAINER_GET_SUBSET (cb.container, &sensor_index);
	  if (array != NULL)
	    {
	      if (array->size > 0)
		{
		  for (i = 0; i < array->size; i++)
		    {
		      ctx = array->array[i];
		      if (ctx->rdr_id == rdr_id)
			{
			  // Found the duplicate entry!
			  sensor_context = ctx;
			  index_oid[1] = ctx->resource_id;
			  index_oid[2] = ctx->saHpiSensorIndex;
			  DEBUGMSGTL ((AGENT,
				       "duplicate sensor entry %d, %d, %d [rdr: %d] found.\n",
				       rpt_entry->DomainId,
				       rpt_entry->ResourceId, sensor->Num,
				       rdr_id));
			  break;
			}
		    }
		}
	      free (array->array);
	      free (array);
	      array = NULL;
	    }
	  // restoree it to its previous glory.
	  sensor_index.len = SENSOR_INDEX_NR;
	}
#endif
      if (!sensor_context)
	{
	  // New entry. Add it
	  sensor_context = saHpiSensorTable_create_row (&sensor_index);
	}
      if (!sensor_context)
	{
	  snmp_log (LOG_ERR, "Not enough memory for a sensor row!");
	  return AGENT_ERR_INTERNAL_ERROR;
	}
      // Generate our full OID
      column[0] = 1;
      column[1] = COLUMN_SAHPISENSORINDEX;

      build_full_oid (saHpiSensorTable_oid, saHpiSensorTable_oid_len,
		      column, 2,
		      &sensor_index, sensor_oid, MAX_OID_LEN, sensor_oid_len);

      // By this stage, sensor_context surely has something in it.
      // '*_modify_context' does a checksum check to see if 
      // the record needs to be altered, and if so populates with
      // information from RDR and the OIDs passed.

      // Get Threshold Data

      rc = getSaHpiSession (&session_id);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		       rc));
	  return rc;
	}
      if (sensor->ThresholdDefn.IsThreshold == SAHPI_TRUE)
	{
	  rc = saHpiSensorThresholdsGet (session_id,
					 rpt_entry->ResourceId,
					 sensor->Num, &sensor_threshold);

	  if (rc != SA_OK)
	    {
	      snmp_log (LOG_ERR,
			"Call to saHpiSensorThresholdsGet fails with return code: %s.\n",
			get_error_string (rc));
	      DEBUGMSGTL ((AGENT,
			   "Call to  SensorThresholdGet fails with return code: %s.\n",
			   get_error_string (rc)));
	      return AGENT_ERR_OPERATION;
	    }
	}
      /*
       * No need to call a plugin that does not support
       * events.
       */
      if (sensor->EventCtrl != SAHPI_SEC_NO_EVENTS)
	{
	  rc = saHpiSensorEventEnablesGet (session_id,
					   rpt_entry->ResourceId,
					   sensor->Num, &enables);

	  if (rc != SA_OK)
	    {
	      snmp_log (LOG_ERR,
			"Call to saHpiSensorEventEnablesGet fails with return code: %s.\n",
			get_error_string (rc));
	      DEBUGMSGTL ((AGENT,
			   "Call to  saHpiSensorEventEnablesGet fails with return code: %s.\n",
			   get_error_string (rc)));
	      // We continue on working. No need to bail on that one - will just use 'undefined(0)' values.
	    }
	}
      rc = saHpiSensorReadingGet (session_id,
				  rpt_entry->ResourceId,
				  sensor->Num, &current_reading);

      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiSensorReadingGet fails with return code: %s.\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to  saHpiSensorReadingGet fails with return code: %s.\n",
		       get_error_string (rc)));
	}

      if (saHpiSensorTable_modify_context (rdr_id, sensor,
					   &enables,
					   rpt_entry,
					   rdr_entry_oid, rdr_entry_oid_len,
					   sensor_context) == AGENT_NEW_ENTRY)
	{
	  CONTAINER_INSERT (cb.container, sensor_context);
	  sensor_count = CONTAINER_SIZE (cb.container);
	}
      /*
         Update our child sensor tables
       */



      rc = populate_ReadingCurrent (rpt_entry->DomainId,
				    rpt_entry->ResourceId,
				    sensor->Num,
				    sensor->Category, &current_reading);

      if (rc != AGENT_ERR_NOERROR)
	DEBUGMSGTL ((AGENT,
		     "call to populate_ReadingCurrent failed with rc: %d\n",
		     rc));


      if (sensor_context->flags & SAHPI_SRF_MIN)
	{
	  rc = populate_ReadingMin (rpt_entry->DomainId,
				    rpt_entry->ResourceId,
				    sensor->Num,
				    sensor->Category,
				    &sensor->DataFormat.Range.Min);
	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to populate_ReadingMin failed with rc: %d\n",
			 rc));
	}
      if (sensor_context->flags & SAHPI_SRF_MAX)
	{
	  rc = populate_ReadingMax (rpt_entry->DomainId,
				    rpt_entry->ResourceId,
				    sensor->Num,
				    sensor->Category,
				    &sensor->DataFormat.Range.Max);
	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to populate_ReadingMax failed with rc: %d\n",
			 rc));
	}

      if (sensor_context->flags & SAHPI_SRF_NOMINAL)
	{
	  rc = populate_ReadingNominal (rpt_entry->DomainId,
					rpt_entry->ResourceId,
					sensor->Num,
					sensor->Category,
					&sensor->DataFormat.Range.Nominal);
	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to populate_ReadingNominal failed with rc: %d\n",
			 rc));
	}
      if (sensor_context->flags & SAHPI_SRF_NORMAL_MAX)
	{
	  rc = populate_ReadingNormalMax (rpt_entry->DomainId,
					  rpt_entry->ResourceId,
					  sensor->Num,
					  sensor->Category,
					  &sensor->DataFormat.Range.
					  NormalMax);
	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to populate_ReadingNormalMax failed with rc: %d\n",
			 rc));
	}


      if (sensor_context->flags & SAHPI_SRF_NORMAL_MIN)
	{
	  rc = populate_ReadingNormalMin (rpt_entry->DomainId,
					  rpt_entry->ResourceId,
					  sensor->Num,
					  sensor->Category,
					  &sensor->DataFormat.Range.
					  NormalMin);
	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to populate_ReadingNormalMin failed with rc: %d\n",
			 rc));
	}


      /*
       * Thresholds
       */
      if (sensor_context->saHpiSensorHasThresholds == MIB_TRUE)
	{
	  /* if (sensor_context->thd_capabilities & SAHPI_STM_LOW_CRIT) */
	  {
	    rc = populate_ThdLowCritical (rpt_entry->DomainId,
					  rpt_entry->ResourceId,
					  sensor->Num,
					  &sensor->ThresholdDefn,
					  &sensor_threshold.LowCritical);
	    if (rc != AGENT_ERR_NOERROR)
	      DEBUGMSGTL ((AGENT,
			   "call to populate_ThdLowCritical failed with rc: %d\n",
			   rc));
	  }

	  /* if (sensor_context->thd_capabilities & SAHPI_STM_LOW_MAJOR) */
	  {
	    rc = populate_ThdLowMajor (rpt_entry->DomainId,
				       rpt_entry->ResourceId,
				       sensor->Num,
				       &sensor->ThresholdDefn,
				       &sensor_threshold.LowMajor);
	    if (rc != AGENT_ERR_NOERROR)
	      DEBUGMSGTL ((AGENT,
			   "call to populate_ThdLowMajor failed with rc: %d\n",
			   rc));
	  }

	  /* if (sensor_context->thd_capabilities & SAHPI_STM_LOW_MINOR) */
	  {
	    rc = populate_ThdLowMinor (rpt_entry->DomainId,
				       rpt_entry->ResourceId,
				       sensor->Num,
				       &sensor->ThresholdDefn,
				       &sensor_threshold.LowMinor);
	    if (rc != AGENT_ERR_NOERROR)
	      DEBUGMSGTL ((AGENT,
			   "call to populate_ThdLowMinor failed with rc: %d\n",
			   rc));
	  }
	  /* if (sensor_context->thd_capabilities & SAHPI_STM_UP_CRIT) */
	  {
	    rc = populate_ThdUpCritical (rpt_entry->DomainId,
					 rpt_entry->ResourceId,
					 sensor->Num,
					 &sensor->ThresholdDefn,
					 &sensor_threshold.UpCritical);
	    if (rc != AGENT_ERR_NOERROR)
	      DEBUGMSGTL ((AGENT,
			   "call to populate_ThdUpCritical failed with rc: %d\n",
			   rc));
	  }

	  /* if (sensor_context->thd_capabilities & SAHPI_STM_UP_MAJOR) */
	  {
	    rc = populate_ThdUpMajor (rpt_entry->DomainId,
				      rpt_entry->ResourceId,
				      sensor->Num,
				      &sensor->ThresholdDefn,
				      &sensor_threshold.UpMajor);
	    if (rc != AGENT_ERR_NOERROR)
	      DEBUGMSGTL ((AGENT,
			   "call to populate_ThdUpMajor failed with rc: %d\n",
			   rc));
	  }

	  /* if (sensor_context->thd_capabilities & SAHPI_STM_UP_MINOR) */
	  {
	    rc = populate_ThdUpMinor (rpt_entry->DomainId,
				      rpt_entry->ResourceId,
				      sensor->Num,
				      &sensor->ThresholdDefn,
				      &sensor_threshold.UpMinor);
	    if (rc != AGENT_ERR_NOERROR)
	      DEBUGMSGTL ((AGENT,
			   "call to populate_ThdUpMinor failed with rc: %d\n",
			   rc));
	  }
	  /* if (sensor_context->thd_capabilities & SAHPI_STM_UP_HYSTERESIS) */
	  {
	    rc = populate_ThdPosHysteresis (rpt_entry->DomainId,
					    rpt_entry->ResourceId,
					    sensor->Num,
					    &sensor->ThresholdDefn,
					    &sensor_threshold.
					    PosThdHysteresis);
	    if (rc != AGENT_ERR_NOERROR)
	      DEBUGMSGTL ((AGENT,
			   "call to populate_ThdPosHysteresis failed with rc: %d\n",
			   rc));
	  }
	  /* if (sensor_context->thd_capabilities & SAHPI_STM_LOW_HYSTERESIS) */
	  {
	    rc = populate_ThdNegHysteresis (rpt_entry->DomainId,
					    rpt_entry->ResourceId,
					    sensor->Num,
					    &sensor->ThresholdDefn,
					    &sensor_threshold.
					    NegThdHysteresis);
	    if (rc != AGENT_ERR_NOERROR)
	      DEBUGMSGTL ((AGENT,
			   "call to populate_ThdNegHysteresis failed with rc: %d\n",
			   rc));
	  }
	}
      rc = AGENT_ERR_NOERROR;
    }
  else
    rc = AGENT_ERR_OPERATION;

  DEBUGMSGTL ((AGENT, "\n\t--- populate_sensor. Exit\n"));
  return rc;
}


int
delete_sensor_row (SaHpiDomainIdT domain_id,
		   SaHpiResourceIdT resource_id, SaHpiSensorNumT sensor_num)
{

  saHpiSensorTable_context *ctx;
  oid index_oid[SENSOR_INDEX_NR];
  netsnmp_index sensor_index;
  int rc = AGENT_ERR_NOT_FOUND;

  DEBUGMSGTL ((AGENT, "delete_sensor_row (%d, %d, %d). Entry.\n",
	       domain_id, resource_id, sensor_num));
  // Look at the MIB to find out what the indexs are
  index_oid[0] = domain_id;
  index_oid[1] = resource_id;
  index_oid[2] = sensor_num;
  // Possible more indexs?
  sensor_index.oids = (oid *) & index_oid;
  sensor_index.len = SENSOR_INDEX_NR;

  ctx = CONTAINER_FIND (cb.container, &sensor_index);

  if (ctx)
    {
      rc = delete_ReadingCurrent_row (ctx->domain_id,
				      ctx->resource_id,
				      ctx->saHpiSensorIndex);


      if (rc != AGENT_ERR_NOERROR)
	DEBUGMSGTL ((AGENT,
		     "call to delete_ReadingCurrent failed with rc: %d\n",
		     rc));

      if (ctx->flags & SAHPI_SRF_MIN)
	{
	  rc = delete_ReadingMin_row (ctx->domain_id,
				      ctx->resource_id,
				      ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ReadingMin failed with rc: %d\n",
			 rc));
	}
      if (ctx->flags & SAHPI_SRF_MAX)
	{
	  rc = delete_ReadingMax_row (ctx->domain_id,
				      ctx->resource_id,
				      ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ReadingMaxfailed with rc: %d\n",
			 rc));
	}
      if (ctx->flags & SAHPI_SRF_NORMAL_MIN)
	{
	  rc = delete_ReadingNormalMin_row (ctx->domain_id,
					    ctx->resource_id,
					    ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ReadingNormalMin failed with rc: %d\n",
			 rc));
	}
      if (ctx->flags & SAHPI_SRF_NORMAL_MAX)
	{
	  rc = delete_ReadingNormalMax_row (ctx->domain_id,
					    ctx->resource_id,
					    ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ReadingNormalMax failed with rc: %d\n",
			 rc));
	}
      if (ctx->flags & SAHPI_SRF_NOMINAL)
	{
	  rc = delete_ReadingNominal_row (ctx->domain_id,
					  ctx->resource_id,
					  ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ReadingNominal failed with rc: %d\n",
			 rc));
	}


      /*
       * Threshold rows
       */
      if (ctx->saHpiSensorHasThresholds == MIB_TRUE)
	{
	  rc = delete_ThdLowCritical_row (ctx->domain_id,
					  ctx->resource_id,
					  ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ThdLowCritical failed with rc: %d\n",
			 rc));

	  rc = delete_ThdLowMajor_row (ctx->domain_id,
				       ctx->resource_id,
				       ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ThdLowMajor failed with rc: %d\n",
			 rc));
	  rc = delete_ThdLowMinor_row (ctx->domain_id,
				       ctx->resource_id,
				       ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ThdLowMinor failed with rc: %d\n",
			 rc));
	  rc = delete_ThdUpCritical_row (ctx->domain_id,
					 ctx->resource_id,
					 ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ThdUpCritical failed with rc: %d\n",
			 rc));
	  rc = delete_ThdUpMajor_row (ctx->domain_id,
				      ctx->resource_id,
				      ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ThdUpMajor failed with rc: %d\n",
			 rc));
	  rc = delete_ThdUpMinor_row (ctx->domain_id,
				      ctx->resource_id,
				      ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ThdUpMinor failed with rc: %d\n",
			 rc));
	  rc = delete_ThdPosHysteresis_row (ctx->domain_id,
					    ctx->resource_id,
					    ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ThdPosHysteresis failed with rc: %d\n",
			 rc));
	  rc = delete_ThdNegHysteresis_row (ctx->domain_id,
					    ctx->resource_id,
					    ctx->saHpiSensorIndex);

	  if (rc != AGENT_ERR_NOERROR)
	    DEBUGMSGTL ((AGENT,
			 "call to delete_ThdNegHysteresis failed with rc: %d\n",
			 rc));
	}
      CONTAINER_REMOVE (cb.container, ctx);
      saHpiSensorTable_delete_row (ctx);
      sensor_count = CONTAINER_SIZE (cb.container);
      rc = AGENT_ERR_NOERROR;
    }
  DEBUGMSGTL ((AGENT, "delete_sensor_row. Exit (rc: %d).\n", rc));
  return rc;
}

int
saHpiSensorTable_modify_context (SaHpiEntryIdT rdr_id,
				 SaHpiSensorRecT * entry,
				 SaHpiSensorEvtEnablesT * enables,
				 SaHpiRptEntryT * rpt_entry,
				 oid * rdr_entry, size_t rdr_entry_oid_len,
				 saHpiSensorTable_context * ctx)
{
  unsigned int update_entry = MIB_FALSE;
  long hash;
  int i, len;
  SaHpiSensorDataFormatT data;


  // Make sure they are valid.
  if (entry && ctx)
    {


      hash = calculate_hash_value (entry, sizeof (SaHpiSensorRecT));

      DEBUGMSGTL ((AGENT, " Hash value: %d, in ctx: %d\n", hash, ctx->hash));

      if (ctx->hash != 0)
	{
	  // Only do the check if the hash value is something else than zero.
	  // 'zero' value is only for newly created records, and in some
	  // rare instances when the hash has rolled to zero - in which
	  // case we will just consider the worst-case scenario and update
	  // the record and not trust the hash value.
	  if (hash == ctx->hash)
	    {
	      // The same data. No need to change.
	      return AGENT_ENTRY_EXIST;
	    }
	  if (((ctx->domain_id == rpt_entry->DomainId) &&
	       (ctx->resource_id == rpt_entry->ResourceId) &&
	       (ctx->saHpiSensorIndex == entry->Num)) ||
	      (ctx->rdr_id == rdr_id))
	    {
	      update_entry = MIB_TRUE;
	      DEBUGMSGTL ((AGENT, "Updating sensor entry. [%d, %d, %d]\n",
			   rpt_entry->DomainId,
			   rpt_entry->ResourceId, entry->Num));
	    }
	}
      if (hash == 0)		// Might happend - roll-over
	hash = 1;		// Need this - we consider hash values of '0' uninitialized
      ctx->hash = hash;

      data = entry->DataFormat;
      ctx->flags = data.Range.Flags;

      ctx->saHpiSensorRDR_len = rdr_entry_oid_len * sizeof (oid);
      memcpy (ctx->saHpiSensorRDR, rdr_entry, ctx->saHpiSensorRDR_len);

      ctx->saHpiSensorIndex = entry->Num;
      ctx->saHpiSensorType = entry->Type + 1;
      ctx->saHpiSensorCategory = entry->Category + 1;
      ctx->saHpiSensorHasThresholds =
	(entry->ThresholdDefn.IsThreshold ==
	 SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

      ctx->saHpiSensorThresholdCapabilities =
	entry->ThresholdDefn.TholdCapabilities + 1;

      // DOAMIN
      ctx->domain_id = rpt_entry->DomainId;
      ctx->resource_id = rpt_entry->ResourceId;
      ctx->rdr_id = rdr_id;
      // IBM-KR: Adding +1
      ctx->saHpiSensorEventsCategoryControl = entry->EventCtrl + 1;

      /*
       * Generate a string representation of the state
       */

      build_state_string (entry->Category,
			  entry->Events,
			  (char *) &ctx->saHpiSensorEventsSupported,
			  &ctx->saHpiSensorEventsSupported_len,
			  SENSOR_EVENTS_SUPPORTED_MAX);
      /*
       * No need to update a sensor that does not support
       * events.
       */  
      if (entry->EventCtrl != SAHPI_SEC_NO_EVENTS)
	{
	  /* we are treating this as an enumeration */
	  ctx->saHpiSensorStatus = enables->SensorStatus +1;

	  build_state_string (entry->Category,
			      enables->AssertEvents,
			      (char *) &ctx->saHpiSensorAssertEvents,
			      &ctx->saHpiSensorAssertEvents_len,
			      SENSOR_EVENTS_SUPPORTED_MAX);

	  build_state_string (entry->Category,
			      enables->DeassertEvents,
			      (char *) &ctx->saHpiSensorDeassertEvents,
			      &ctx->saHpiSensorDeassertEvents_len,
			      SENSOR_EVENTS_SUPPORTED_MAX);
	}


      ctx->saHpiSensorIgnore =
	(entry->Ignore == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

      ctx->saHpiSensorReadingFormats = data.ReadingFormats;
      ctx->saHpiSensorIsNumeric =
	(data.IsNumeric == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

      ctx->saHpiSensorSignFormat = data.SignFormat + 1;
      ctx->saHpiSensorBaseUnits = data.BaseUnits + 1;
      ctx->saHpiSensorModifierUnits = data.ModifierUnits + 1;
      ctx->saHpiSensorModifierUse = data.ModifierUse + 1;
      ctx->saHpiSensorFactorsStatic =
	(data.FactorsStatic == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

      // Sensor factors.
      memset (ctx->saHpiSensorFactors, 0x00, SAHPISENSORFACTORS_MAX);
      // Normalize data.
      data.Factors.M_Factor = htons (data.Factors.M_Factor);
      data.Factors.B_Factor = htons (data.Factors.B_Factor);
      data.Factors.AccuracyFactor = htons (data.Factors.AccuracyFactor);

      memcpy (ctx->saHpiSensorFactors, &data.Factors.M_Factor,
	      sizeof (SaHpiInt16T));
      memcpy (ctx->saHpiSensorFactors + sizeof (SaHpiInt16T),
	      &data.Factors.B_Factor, sizeof (SaHpiInt16T));

      memcpy (ctx->saHpiSensorFactors + (sizeof (SaHpiInt16T) * 2),
	      &data.Factors.AccuracyFactor, sizeof (SaHpiUint16T));

      ctx->saHpiSensorFactors[6] = data.Factors.ToleranceFactor;
      ctx->saHpiSensorFactors[7] = data.Factors.ExpA;
      ctx->saHpiSensorFactors[8] = data.Factors.ExpR;
      ctx->saHpiSensorFactors[9] = data.Factors.ExpB;
      ctx->saHpiSensorFactors_len = SAHPISENSORFACTORS_MAX;

      ctx->saHpiSensorFactorsLinearization = data.Factors.Linearization + 1;

      ctx->saHpiSensorPercentage =
	(data.Percentage == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;

      ctx->saHpiSensorOEM = entry->Oem;

      ctx->saHpiSensorRangeFlags_len = 0;

      for (i = 0; i < RANGE_FLAGS_LEN; i++)
	{
	  if ((data.Range.Flags & range_flags[i].flag) == range_flags[i].flag)
	    {
	      len = strlen (range_flags[i].str);
	      if (len + ctx->saHpiSensorRangeFlags_len >
		  SENSOR_RANGE_FLAGS_MAX)
		break;

	      memcpy (ctx->saHpiSensorRangeFlags +
		      ctx->saHpiSensorRangeFlags_len, range_flags[i].str,
		      len);

	      ctx->saHpiSensorRangeFlags_len =
		ctx->saHpiSensorRangeFlags_len + len;

	    }

	}
      ctx->saHpiSensorRangeFlags_len -= 2;
      if (update_entry == MIB_TRUE)
	return AGENT_ENTRY_EXIST;
      return AGENT_NEW_ENTRY;
    }

  return AGENT_ERR_NULL_DATA;
}

int
set_sensor_event (saHpiSensorTable_context * ctx)
{


  SaHpiSessionIdT session_id;
  SaErrorT rc;
  SaHpiSensorEvtEnablesT enables;

  if (ctx)
    {
      memset (&enables, 0x00, sizeof (SaHpiSensorEvtEnablesT));
      /* we are treating this as an enumeration */
      enables.SensorStatus = ctx->saHpiSensorStatus -1;

      rc = build_state_value (ctx->saHpiSensorAssertEvents,
			      ctx->saHpiSensorAssertEvents_len,
			      &enables.AssertEvents);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT,
		       "Call to build_state_value with [%s] failed with rc: %d.\n",
		       ctx->saHpiSensorAssertEvents, rc));
	  return AGENT_ERR_WRONG_DELIM;
	}
      rc = build_state_value (ctx->saHpiSensorDeassertEvents,
			      ctx->saHpiSensorDeassertEvents_len,
			      &enables.DeassertEvents);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT,
		       "Call to build_state_value with [%s] failed with rc: %d.\n",
		       ctx->saHpiSensorDeassertEvents, rc));
	  return AGENT_ERR_WRONG_DELIM;
	}
      DEBUGMSGTL ((AGENT,
		   "enables.SensorStatus:   %X\n"
		   "enables.AssertEvents:   %X[%s]\n"
		   "enables.DeassertEvents: %X[%s]\n",
		   enables.SensorStatus, enables.AssertEvents,
		   ctx->saHpiSensorAssertEvents,
		   enables.DeassertEvents, ctx->saHpiSensorDeassertEvents));
      // Get the seesion_id
      rc = getSaHpiSession (&session_id);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		       rc));
	  return rc;
	}
      DEBUGMSGTL ((AGENT, "Calling 'saHpiSensorEventEnablesSet'\n"));
      rc = saHpiSensorEventEnablesSet (session_id,
				       ctx->resource_id,
				       ctx->saHpiSensorIndex, &enables);

      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiSensorEventEnablesSet failed wit return code %s.\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to saHpiSensorEventEnablesSet failed wit return code %s.\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}

      build_state_string (ctx->saHpiSensorCategory - 1,
			  enables.AssertEvents,
			  (char *) &ctx->saHpiSensorAssertEvents,
			  &ctx->saHpiSensorAssertEvents_len,
			  SENSOR_EVENTS_SUPPORTED_MAX);

      build_state_string (ctx->saHpiSensorCategory - 1,
			  enables.DeassertEvents,
			  (char *) &ctx->saHpiSensorDeassertEvents,
			  &ctx->saHpiSensorDeassertEvents_len,
			  SENSOR_EVENTS_SUPPORTED_MAX);

      return AGENT_ERR_NOERROR;
    }
  return AGENT_ERR_NULL_DATA;
}


/************************************************************
 * the *_row_copy routine
 */
static int
saHpiSensorTable_row_copy (saHpiSensorTable_context * dst,
			   saHpiSensorTable_context * src)
{


  if (!dst || !src)
    return 1;

  /*
   * copy index, if provided
   */
  if (dst->index.oids)
    free (dst->index.oids);
  if (snmp_clone_mem ((void *) &dst->index.oids, src->index.oids,
		      src->index.len * sizeof (oid)))
    {
      dst->index.oids = NULL;
      return 1;
    }
  dst->index.len = src->index.len;


  /*
   * copy components into the context structure
   */
  dst->saHpiSensorIndex = src->saHpiSensorIndex;

  dst->saHpiSensorType = src->saHpiSensorType;

  dst->saHpiSensorCategory = src->saHpiSensorCategory;

  dst->saHpiSensorEventsCategoryControl =
    src->saHpiSensorEventsCategoryControl;

  memcpy (dst->saHpiSensorEventsSupported,
	  src->saHpiSensorEventsSupported,
	  src->saHpiSensorEventsSupported_len);
  dst->saHpiSensorEventsSupported_len = src->saHpiSensorEventsSupported_len;

  dst->saHpiSensorStatus = src->saHpiSensorStatus;

  memcpy (dst->saHpiSensorAssertEvents, src->saHpiSensorAssertEvents,
	  src->saHpiSensorAssertEvents_len);
  dst->saHpiSensorAssertEvents_len = src->saHpiSensorAssertEvents_len;

  memcpy (dst->saHpiSensorDeassertEvents, src->saHpiSensorDeassertEvents,
	  src->saHpiSensorDeassertEvents_len);
  dst->saHpiSensorDeassertEvents_len = src->saHpiSensorDeassertEvents_len;

  dst->saHpiSensorIgnore = src->saHpiSensorIgnore;

  dst->saHpiSensorReadingFormats = src->saHpiSensorReadingFormats;

  dst->saHpiSensorIsNumeric = src->saHpiSensorIsNumeric;

  dst->saHpiSensorSignFormat = src->saHpiSensorSignFormat;

  dst->saHpiSensorBaseUnits = src->saHpiSensorBaseUnits;

  dst->saHpiSensorModifierUnits = src->saHpiSensorModifierUnits;

  dst->saHpiSensorModifierUse = src->saHpiSensorModifierUse;

  dst->saHpiSensorFactorsStatic = src->saHpiSensorFactorsStatic;

  memcpy (dst->saHpiSensorFactors, src->saHpiSensorFactors,
	  src->saHpiSensorFactors_len);
  dst->saHpiSensorFactors_len = src->saHpiSensorFactors_len;

  dst->saHpiSensorFactorsLinearization = src->saHpiSensorFactorsLinearization;

  dst->saHpiSensorPercentage = src->saHpiSensorPercentage;

  memcpy (dst->saHpiSensorRangeFlags, src->saHpiSensorRangeFlags,
	  src->saHpiSensorRangeFlags_len);
  dst->saHpiSensorRangeFlags_len = src->saHpiSensorRangeFlags_len;

  dst->saHpiSensorHasThresholds = src->saHpiSensorHasThresholds;

  dst->saHpiSensorThresholdCapabilities =
    src->saHpiSensorThresholdCapabilities;

  dst->saHpiSensorOEM = src->saHpiSensorOEM;

  memcpy (dst->saHpiSensorRDR, src->saHpiSensorRDR, src->saHpiSensorRDR_len);
  dst->saHpiSensorRDR_len = src->saHpiSensorRDR_len;

  dst->flags = src->flags;
  dst->rdr_id = src->rdr_id;
  dst->resource_id = src->resource_id;
  dst->domain_id = src->domain_id;
  dst->hash = src->hash;

  return 0;
}



/*
 * the *_extract_index routine
 */
int
saHpiSensorTable_extract_index (saHpiSensorTable_context * ctx,
				netsnmp_index * hdr)
{
  /*
   * temporary local storage for extracting oid index
   */
  netsnmp_variable_list var_saHpiDomainID;
  netsnmp_variable_list var_saHpiResourceID;

  netsnmp_variable_list var_saHpiSensorIndex;
  int err;

  /*
   * copy index, if provided
   */
  if (hdr)
    {
      netsnmp_assert (ctx->index.oids == NULL);
      if (snmp_clone_mem ((void *) &ctx->index.oids, hdr->oids,
			  hdr->len * sizeof (oid)))
	{
	  return -1;
	}
      ctx->index.len = hdr->len;
    }

    /**
     * Create variable to hold each component of the index
     */
  memset (&var_saHpiDomainID, 0x00, sizeof (var_saHpiDomainID));
  var_saHpiDomainID.type = ASN_UNSIGNED;
  var_saHpiDomainID.next_variable = &var_saHpiResourceID;

  memset (&var_saHpiResourceID, 0x00, sizeof (var_saHpiResourceID));
  var_saHpiResourceID.type = ASN_UNSIGNED;
  var_saHpiResourceID.next_variable = &var_saHpiSensorIndex;

  memset (&var_saHpiSensorIndex, 0x00, sizeof (var_saHpiSensorIndex));
  var_saHpiSensorIndex.type = ASN_UNSIGNED;
  var_saHpiSensorIndex.next_variable = NULL;

  /*
   * parse the oid into the individual components
   */
  err = parse_oid_indexes (hdr->oids, hdr->len, &var_saHpiDomainID);
  if (err == SNMP_ERR_NOERROR)
    {
      ctx->saHpiSensorIndex = *var_saHpiSensorIndex.val.integer;
    }

  /*
   * parsing may have allocated memory. free it.
   */
  snmp_reset_var_buffers (&var_saHpiDomainID);

  return err;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int
saHpiSensorTable_can_delete (saHpiSensorTable_context * undo_ctx,
			     saHpiSensorTable_context * row_ctx,
			     netsnmp_request_group * rg)
{
  return 1;
}


/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiSensorTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiSensorTable_context *
saHpiSensorTable_create_row (netsnmp_index * hdr)
{
  saHpiSensorTable_context *ctx =
    SNMP_MALLOC_TYPEDEF (saHpiSensorTable_context);
  if (!ctx)
    return NULL;

  if (saHpiSensorTable_extract_index (ctx, hdr))
    {
      free (ctx->index.oids);
      free (ctx);
      return NULL;
    }
  ctx->saHpiSensorStatus = 0;
  ctx->saHpiSensorAssertEvents_len = 0;
  ctx->saHpiSensorDeassertEvents_len = 0;
  ctx->hash = 0;
  ctx->saHpiSensorHasThresholds = MIB_FALSE;
  ctx->saHpiSensorThresholdCapabilities = 0;
  return ctx;
}


/************************************************************
 * the *_duplicate row routine
 */
saHpiSensorTable_context *
saHpiSensorTable_duplicate_row (saHpiSensorTable_context * row_ctx)
{
  saHpiSensorTable_context *dup;
  if (!row_ctx)
    return NULL;

  dup = SNMP_MALLOC_TYPEDEF (saHpiSensorTable_context);
  if (!dup)
    return NULL;

  if (saHpiSensorTable_row_copy (dup, row_ctx))
    {
      free (dup);
      dup = NULL;
    }

  return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index *
saHpiSensorTable_delete_row (saHpiSensorTable_context * ctx)
{
  if (ctx->index.oids)
    free (ctx->index.oids);


  free (ctx);

  return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void
saHpiSensorTable_set_reserve1 (netsnmp_request_group * rg)
{

  netsnmp_variable_list *var;
  netsnmp_request_group_item *current;
  saHpiSensorTable_context *row_ctx =
    (saHpiSensorTable_context *) rg->existing_row;
  int rc = SNMP_ERR_NOERROR;

  DEBUGMSGTL ((AGENT, "saHpiSensorTable_set_reserve1. Entry\n"));
  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;
      rc = SNMP_ERR_NOERROR;

      switch (current->tri->colnum)
	{

	case COLUMN_SAHPISENSORSTATUS:
	    /** ENUM requires ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiSensorStatus));
	  break;

	case COLUMN_SAHPISENSORASSERTEVENTS:
	  /** OCTETSTR = ASN_OCTET_STR */
	  if (var->type != ASN_OCTET_STR)
	    {
	      rc = SNMP_ERR_WRONGTYPE;
	    }
	  else if (var->val_len > SENSOR_EVENTS_SUPPORTED_MAX)
	    {
	      rc = SNMP_ERR_WRONGLENGTH;
	    }
	  else if (var->val_len < 0)
	    {
	      rc = SNMP_ERR_WRONGLENGTH;
	    }
	  break;

	case COLUMN_SAHPISENSORDEASSERTEVENTS:
	  /** OCTETSTR = ASN_OCTET_STR */
	  if (var->type != ASN_OCTET_STR)
	    {
	      rc = SNMP_ERR_WRONGTYPE;
	    }
	  else if (var->val_len > SENSOR_EVENTS_SUPPORTED_MAX)
	    {
	      rc = SNMP_ERR_WRONGLENGTH;
	    }
	  else if (var->val_len < 0)
	    {
	      rc = SNMP_ERR_WRONGLENGTH;
	    }
	  break;

	case COLUMN_SAHPISENSORINDEX:
	case COLUMN_SAHPISENSORTYPE:
	case COLUMN_SAHPISENSORCATEGORY:
	case COLUMN_SAHPISENSOREVENTSCATEGORYCONTROL:
	case COLUMN_SAHPISENSOREVENTSSUPPORTED:
	case COLUMN_SAHPISENSORIGNORE:
	case COLUMN_SAHPISENSORREADINGFORMATS:
	case COLUMN_SAHPISENSORISNUMERIC:
	case COLUMN_SAHPISENSORSIGNFORMAT:
	case COLUMN_SAHPISENSORBASEUNITS:
	case COLUMN_SAHPISENSORMODIFIERUNITS:
	case COLUMN_SAHPISENSORMODIFIERUSE:
	case COLUMN_SAHPISENSORFACTORSSTATIC:
	case COLUMN_SAHPISENSORFACTORS:
	case COLUMN_SAHPISENSORFACTORSLINEARIZATION:
	case COLUMN_SAHPISENSORPERCENTAGE:
	case COLUMN_SAHPISENSORRANGEFLAGS:
	case COLUMN_SAHPISENSORHASTHRESHOLDS:
	case COLUMN_SAHPISENSORTHRESHOLDCAPABILITIES:
	case COLUMN_SAHPISENSOROEM:
	case COLUMN_SAHPISENSORRDR:
	  rc = SNMP_ERR_NOTWRITABLE;
	  break;
	default:		// Trying to write to read-only objects or to unknown columns?
	  rc = SNMP_ERR_GENERR;
	  break;
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
      rg->status = SNMP_MAX (rg->status, current->ri->status);
    }
  for (current = rg->list; current; current = current->next)
    {

      // The nice thing about this API is that _row_copy() is called
      // for this row - if the API has matched the index with an
      // already existing entry. We check the 'hash' value. If its
      // 0 the API couldn't find the right context.

      if (((saHpiSensorTable_context *) rg->existing_row)->hash == 0)
	{
	  rc = SNMP_ERR_NOSUCHNAME;
	  netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
	}
    }
  DEBUGMSGTL ((AGENT, "saHpiSensorTable_set_reserve1: Exit (rc:%d)\n", rc));
}

void
saHpiSensorTable_set_reserve2 (netsnmp_request_group * rg)
{
  netsnmp_request_group_item *current;
  netsnmp_variable_list *var;
  int rc = SNMP_ERR_NOERROR;

  DEBUGMSGTL ((AGENT, "saHpiSensorTable_set_reserve2: Entry\n"));
  rg->rg_void = rg->list->ri;

  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;

      switch (current->tri->colnum)
	{
      case COLUMN_SAHPISENSORSTATUS:             
          *var->val.integer --;
          if ((*var->val.integer != SAHPI_SENSTAT_EVENTS_ENABLED) &&
              (*var->val.integer != SAHPI_SENSTAT_SCAN_ENABLED) &&
              (*var->val.integer != SAHPI_SENSTAT_BUSY) &&
              (*var->val.integer != (SAHPI_SENSTAT_EVENTS_ENABLED &
	       SAHPI_SENSTAT_SCAN_ENABLED)) &&
              (*var->val.integer != SAHPI_SENSTAT_SCAN_ENABLED &
	       SAHPI_SENSTAT_BUSY) &&
              (*var->val.integer != (SAHPI_SENSTAT_BUSY &
	       SAHPI_SENSTAT_EVENTS_ENABLED)) &&
              (*var->val.integer != (SAHPI_SENSTAT_BUSY &
	       SAHPI_SENSTAT_EVENTS_ENABLED & SAHPI_SENSTAT_SCAN_ENABLED)))
	  {              
		  rc= SNMP_ERR_BADVALUE;
	  }
           *var->val.integer++; 
/*	
	case COLUMN_SAHPISENSORSTATUS:
	  if ((*var->val.integer != SAHPI_SENSTAT_EVENTS_ENABLED) &&
	      (*var->val.integer != SAHPI_SENSTAT_SCAN_ENABLED) &&
	      (*var->val.integer != SAHPI_SENSTAT_BUSY))
	    {
	      rc = SNMP_ERR_BADVALUE;
	    }
*/	    
	  break;
	case COLUMN_SAHPISENSORASSERTEVENTS:
	case COLUMN_SAHPISENSORDEASSERTEVENTS:
	  break;
	default:
		/** We shouldn't get here */
	  netsnmp_assert (0);  /** why wasn't this caught in reserve1? */
	  break;
	}


      if (rc)
	{
	  netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
	}
    }
  DEBUGMSGTL((AGENT,"saHpiTable_reserve2. Exit (rc: %d)", rc));
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void
saHpiSensorTable_set_action (netsnmp_request_group * rg)
{
  netsnmp_variable_list *var;
  saHpiSensorTable_context *row_ctx =
    (saHpiSensorTable_context *) rg->existing_row;

  netsnmp_request_group_item *current;
  int rc = SNMP_ERR_NOERROR;
  int rc2;
  DEBUGMSGTL ((AGENT, "saHpiSensorTable_set_action: Entry\n"));
  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;

      switch (current->tri->colnum)
	{
	case COLUMN_SAHPISENSORSTATUS:
	  /** ENUM requires ASN_INTEGER */
	  row_ctx->saHpiSensorStatus = *var->val.integer;
	  rc2 = set_sensor_event (row_ctx);
	  if (rc2 != AGENT_ERR_NOERROR)
	    rc = SNMP_ERR_GENERR;
	  if (rc2 == AGENT_ERR_WRONG_DELIM)
	    rc = SNMP_ERR_BADVALUE;
	  break;

	case COLUMN_SAHPISENSORASSERTEVENTS:
	  memcpy (row_ctx->saHpiSensorAssertEvents,
		  var->val.string, var->val_len);
	  row_ctx->saHpiSensorAssertEvents_len = var->val_len;
	  rc2 = set_sensor_event (row_ctx);
	  if (rc2 != AGENT_ERR_NOERROR)
	    rc = SNMP_ERR_GENERR;
	  if (rc2 == AGENT_ERR_WRONG_DELIM)
	    rc = SNMP_ERR_BADVALUE;
	  break;

	case COLUMN_SAHPISENSORDEASSERTEVENTS:
	  memcpy (row_ctx->saHpiSensorDeassertEvents,
		  var->val.string, var->val_len);
	  row_ctx->saHpiSensorDeassertEvents_len = var->val_len;
	  rc2 = set_sensor_event (row_ctx);
	  if (rc2 != AGENT_ERR_NOERROR)
	    rc = SNMP_ERR_GENERR;
	  if (rc2 == AGENT_ERR_WRONG_DELIM)
	    rc = SNMP_ERR_BADVALUE;
	  break;

	default:
		/** We shouldn't get here */
	  netsnmp_assert (0);  /** why wasn't this caught in reserve1? */
	  break;
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);

    }
  DEBUGMSGTL ((AGENT, "saHpiSensorTable_set_action: Exit \n"));

}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void
saHpiSensorTable_set_commit (netsnmp_request_group * rg)
{
}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void
saHpiSensorTable_set_free (netsnmp_request_group * rg)
{
}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void
saHpiSensorTable_set_undo (netsnmp_request_group * rg)
{
}




/************************************************************
 *
 * Initialize the saHpiSensorTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiSensorTable (void)
{
  netsnmp_table_registration_info *table_info;

  if (my_handler)
    {
      snmp_log (LOG_ERR,
		"initialize_table_saHpiSensorTable_handler called again\n");
      return;
    }

  memset (&cb, 0x00, sizeof (cb));

    /** create the table structure itself */
  table_info = SNMP_MALLOC_TYPEDEF (netsnmp_table_registration_info);

  /*
   * if your table is read only, it's easiest to change the
   * HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY 
   */
  my_handler = netsnmp_create_handler_registration ("saHpiSensorTable",
						    netsnmp_table_array_helper_handler,
						    saHpiSensorTable_oid,
						    saHpiSensorTable_oid_len,
						    HANDLER_CAN_RWRITE);

  if (!my_handler || !table_info)
    {
      snmp_log (LOG_ERR, "malloc failed in "
		"initialize_table_saHpiSensorTable_handler\n");
      return;	/** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */

  /*
   * internal indexes
   */
	/** index: saHpiSensorIndex */
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);

  table_info->min_column = saHpiSensorTable_COL_MIN;
  table_info->max_column = saHpiSensorTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
  cb.get_value = saHpiSensorTable_get_value;

  cb.container = netsnmp_container_find ("saHpiSensorTable_primary:"
					 "saHpiSensorTable:"
					 "table_container");



  //    cb.can_set = 1;

  cb.create_row = (UserRowMethod *) saHpiSensorTable_create_row;

  cb.duplicate_row = (UserRowMethod *) saHpiSensorTable_duplicate_row;
  cb.delete_row = (UserRowMethod *) saHpiSensorTable_delete_row;
  cb.row_copy = (Netsnmp_User_Row_Operation *) saHpiSensorTable_row_copy;

  cb.can_delete = (Netsnmp_User_Row_Action *) saHpiSensorTable_can_delete;

  cb.set_reserve1 = saHpiSensorTable_set_reserve1;
  cb.set_reserve2 = saHpiSensorTable_set_reserve2;
  cb.set_action = saHpiSensorTable_set_action;
  cb.set_commit = saHpiSensorTable_set_commit;
  cb.set_free = saHpiSensorTable_set_free;
  cb.set_undo = saHpiSensorTable_set_undo;

  DEBUGMSGTL (("initialize_table_saHpiSensorTable",
	       "Registering table saHpiSensorTable " "as a table array\n"));

  netsnmp_table_container_register (my_handler, table_info, &cb,
				    cb.container, 1);


  netsnmp_register_read_only_counter32_instance ("sensor_count",
						 saHpiSensorCount_oid,
						 OID_LENGTH
						 (saHpiSensorCount_oid),
						 &sensor_count, NULL);
}


/************************************************************
 * saHpiSensorTable_get_value
 */
int
saHpiSensorTable_get_value (netsnmp_request_info * request,
			    netsnmp_index * item,
			    netsnmp_table_request_info * table_info)
{
  netsnmp_variable_list *var = request->requestvb;
  saHpiSensorTable_context *context = (saHpiSensorTable_context *) item;

  switch (table_info->colnum)
    {

    case COLUMN_SAHPISENSORINDEX:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->saHpiSensorIndex,
				sizeof (context->saHpiSensorIndex));
      break;

    case COLUMN_SAHPISENSORTYPE:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorType,
				sizeof (context->saHpiSensorType));
      break;

    case COLUMN_SAHPISENSORCATEGORY:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorCategory,
				sizeof (context->saHpiSensorCategory));
      break;

    case COLUMN_SAHPISENSOREVENTSCATEGORYCONTROL:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorEventsCategoryControl,
				sizeof (context->
					saHpiSensorEventsCategoryControl));
      break;

    case COLUMN_SAHPISENSOREVENTSSUPPORTED:
	    /** OCTETSTR = ASN_OCTET_STR */
      snmp_set_var_typed_value (var, ASN_OCTET_STR,
				(char *) &context->
				saHpiSensorEventsSupported,
				context->saHpiSensorEventsSupported_len);
      break;

    case COLUMN_SAHPISENSORSTATUS:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorStatus,
				sizeof (context->saHpiSensorStatus));
      break;

    case COLUMN_SAHPISENSORASSERTEVENTS:
	    /** OCTETSTR = ASN_OCTET_STR */
      snmp_set_var_typed_value (var, ASN_OCTET_STR,
				(char *) &context->
				saHpiSensorAssertEvents,
				context->saHpiSensorAssertEvents_len);
      break;

    case COLUMN_SAHPISENSORDEASSERTEVENTS:
	    /** OCTETSTR = ASN_OCTET_STR */
      snmp_set_var_typed_value (var, ASN_OCTET_STR,
				(char *) &context->
				saHpiSensorDeassertEvents,
				context->saHpiSensorDeassertEvents_len);
      break;

    case COLUMN_SAHPISENSORIGNORE:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorIgnore,
				sizeof (context->saHpiSensorIgnore));
      break;

    case COLUMN_SAHPISENSORREADINGFORMATS:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->
				saHpiSensorReadingFormats,
				sizeof (context->saHpiSensorReadingFormats));
      break;

    case COLUMN_SAHPISENSORISNUMERIC:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorIsNumeric,
				sizeof (context->saHpiSensorIsNumeric));
      break;

    case COLUMN_SAHPISENSORSIGNFORMAT:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorSignFormat,
				sizeof (context->saHpiSensorSignFormat));
      break;

    case COLUMN_SAHPISENSORBASEUNITS:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorBaseUnits,
				sizeof (context->saHpiSensorBaseUnits));
      break;

    case COLUMN_SAHPISENSORMODIFIERUNITS:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorModifierUnits,
				sizeof (context->saHpiSensorModifierUnits));
      break;

    case COLUMN_SAHPISENSORMODIFIERUSE:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorModifierUse,
				sizeof (context->saHpiSensorModifierUse));
      break;

    case COLUMN_SAHPISENSORFACTORSSTATIC:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorFactorsStatic,
				sizeof (context->saHpiSensorFactorsStatic));
      break;

    case COLUMN_SAHPISENSORFACTORS:
	    /** OCTETSTR = ASN_OCTET_STR */
      snmp_set_var_typed_value (var, ASN_OCTET_STR,
				(char *) &context->saHpiSensorFactors,
				context->saHpiSensorFactors_len);
      break;

    case COLUMN_SAHPISENSORFACTORSLINEARIZATION:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorFactorsLinearization,
				sizeof (context->
					saHpiSensorFactorsLinearization));
      break;

    case COLUMN_SAHPISENSORPERCENTAGE:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiSensorPercentage,
				sizeof (context->saHpiSensorPercentage));
      break;
    case COLUMN_SAHPISENSORRANGEFLAGS:
	    /** OCTETSTR = ASN_OCTET_STR */
      snmp_set_var_typed_value (var, ASN_OCTET_STR,
				(char *) &context->saHpiSensorRangeFlags,
				context->saHpiSensorRangeFlags_len);
      break;
    case COLUMN_SAHPISENSORHASTHRESHOLDS:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorHasThresholds,
				sizeof (context->saHpiSensorHasThresholds));
      break;

    case COLUMN_SAHPISENSORTHRESHOLDCAPABILITIES:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorThresholdCapabilities,
				sizeof (context->
					saHpiSensorThresholdCapabilities));
      break;


    case COLUMN_SAHPISENSOROEM:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->saHpiSensorOEM,
				sizeof (context->saHpiSensorOEM));
      break;

    case COLUMN_SAHPISENSORRDR:
	    /** RowPointer = ASN_OBJECT_ID */
      snmp_set_var_typed_value (var, ASN_OBJECT_ID,
				(char *) &context->saHpiSensorRDR,
				context->saHpiSensorRDR_len);
      break;

    default:
	    /** We shouldn't get here */
      snmp_log (LOG_ERR, "unknown column in " "saHpiSensorTable_get_value\n");
      return SNMP_ERR_GENERR;
    }
  return SNMP_ERR_NOERROR;
}
