/*
 * (C) Copyright IBM Corp. 2004
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   Konrad Rzeszutek <konradr@us.ibm.com>
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.18 2003/11/07 17:03:52 rstory Exp $
 *
 * $Id: saHpiSensorThdUpMinorTable.c,v 1.2 2004/03/02 23:04:06 konradrzeszutek Exp $
 *
 *
 */
#include <SaHpi.h>
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <saHpiSensorThdUpMinorTable.h>


#include <hpiSubagent.h>


static netsnmp_handler_registration *my_handler = NULL;
static netsnmp_table_array_callbacks cb;

oid saHpiSensorThdUpMinorTable_oid[] =
  { saHpiSensorThdUpMinorTable_TABLE_OID };
size_t saHpiSensorThdUpMinorTable_oid_len =
OID_LENGTH (saHpiSensorThdUpMinorTable_oid);

static int
modify_saHpiSensorThdUpMinorTable_row (SaHpiDomainIdT domain_id,
				       SaHpiResourceIdT resource_id,
				       SaHpiSensorNumT sensor_num,
				       SaHpiSensorThdDefnT * threshold_def,
				       SaHpiSensorReadingT * reading,
				       saHpiSensorThdUpMinorTable_context
				       * ctx);

int
delete_ThdUpMinor_row (SaHpiDomainIdT domain_id,
		       SaHpiResourceIdT resource_id,
		       SaHpiSensorNumT sensor_num)
{

  saHpiSensorThdUpMinorTable_context *ctx;
  oid index_oid[SENSOR_THD_INDEX_NR];
  netsnmp_index sensor_reading_index;
  int rc = AGENT_ERR_NOT_FOUND;

  DEBUGMSGTL ((AGENT, "delete_ThdUpMinor_row (%d, %d, %d). Entry.\n",
	       domain_id, resource_id, sensor_num));

  // Look at the MIB to find out what the indexs are
  index_oid[0] = domain_id;
  index_oid[1] = resource_id;
  index_oid[2] = sensor_num;
  // Possible more indexs?
  sensor_reading_index.oids = (oid *) & index_oid;
  sensor_reading_index.len = SENSOR_THD_INDEX_NR;

  ctx = CONTAINER_FIND (cb.container, &sensor_reading_index);

  if (ctx)
    {
      CONTAINER_REMOVE (cb.container, ctx);
      saHpiSensorThdUpMinorTable_delete_row (ctx);
      rc = AGENT_ERR_NOERROR;
    }
  DEBUGMSGTL ((AGENT, "delete_ThdUpMinor_row. Exit (rc: %d).\n", rc));
  return rc;
}


int
populate_ThdUpMinor (SaHpiDomainIdT domain_id,
		     SaHpiResourceIdT resource_id,
		     SaHpiSensorNumT sensor_id,
		     SaHpiSensorThdDefnT * threshold_def,
		     SaHpiSensorReadingT * reading)
{


  int rc = AGENT_ERR_NOERROR;

  oid index_oid[SENSOR_THD_INDEX_NR];
  netsnmp_index sensor_thd_index;
  saHpiSensorThdUpMinorTable_context *ctx;


  DEBUGMSGT ((AGENT, "\t--- populate_ThdUpMinor, Entry,\n"));

  if (reading && threshold_def)
    {

      /*
         Construct index 
       */
      sensor_thd_index.len = SENSOR_THD_INDEX_NR;
      index_oid[0] = domain_id;
      index_oid[1] = resource_id;
      index_oid[2] = sensor_id;
      sensor_thd_index.oids = (oid *) & index_oid;

      /*
       * Find the row.
       */
      ctx = NULL;
      ctx = CONTAINER_FIND (cb.container, &sensor_thd_index);

      if (!ctx)
	{
	  /* 
	     New entry. Create it.
	   */
	  ctx = saHpiSensorThdUpMinorTable_create_row (&sensor_thd_index);
	}
      if (!ctx)
	{
	  snmp_log (LOG_ERR,
		    "Not enough memory for a sensor ThdUpMinor row!");
	  return AGENT_ERR_INTERNAL_ERROR;
	}

      if (modify_saHpiSensorThdUpMinorTable_row (domain_id,
						 resource_id,
						 sensor_id,
						 threshold_def,
						 reading,
						 ctx) == AGENT_NEW_ENTRY)
	{
	  /*
	   * Add new entry.
	   */
	  CONTAINER_INSERT (cb.container, ctx);
	}
      rc = AGENT_ERR_NOERROR;

    }
  else
    rc = AGENT_ERR_OPERATION;
  DEBUGMSGT ((AGENT, "\t--- populate_ThdUpMinor, Exit (rc: %d),\n", rc));

  return rc;
}


int
modify_saHpiSensorThdUpMinorTable_row (SaHpiDomainIdT domain_id,
				       SaHpiResourceIdT resource_id,
				       SaHpiSensorNumT sensor_num,
				       SaHpiSensorThdDefnT * threshold_def,
				       SaHpiSensorReadingT * reading,
				       saHpiSensorThdUpMinorTable_context
				       * ctx)
{

  long hash = 0;
  unsigned int update_entry = MIB_FALSE;
  //  char format[SENSOR_THD_INTER_MAX];

  DEBUGMSGTL ((AGENT, "Modify saHpiSensorThdUpMinorTable_ctx: Entry.\n"));

  if (ctx)
    {
      hash = calculate_hash_value (reading, sizeof (SaHpiSensorReadingT));

      DEBUGMSGTL ((AGENT, " Hash value: %d, in ctx: %d\n", hash, ctx->hash));

      if (ctx->hash != 0)
	{
	  /* Only do the check if the hash value is something else than zero.
	   * 'zero' value is only for newly created records, and in some
	   * rare instances when the hash has rolled to zero - in which
	   * case we will just consider the worst-case scenario and update
	   * the record and not trust the hash value.
	   */
	  if (hash == ctx->hash)
	    {
	      /* The same data. No need to change. */
	      return AGENT_ENTRY_EXIST;
	    }
	  if ((ctx->domain_id == domain_id) &&
	      (ctx->resource_id == resource_id) &&
	      (ctx->sensor_id == sensor_num))
	    {
	      update_entry = MIB_TRUE;
	      DEBUGMSGTL ((AGENT,
			   "Updating ThdUpMinorTable row [%d, %d, %d]\n",
			   domain_id, resource_id, sensor_num));
	    }
	}

      if (hash == 0)		/* Might happend - roll-over */
	hash = 1;		/* Need this - we consider hash
				 * values of '0' uninitialized */
      ctx->hash = hash;
      ctx->resource_id = resource_id;
      ctx->domain_id = domain_id;
      ctx->sensor_id = sensor_num;


      build_reading_strings (reading,
			     0,
			     &ctx->saHpiSensorThdUpMinorValuesPresent,
			     &ctx->saHpiSensorThdUpMinorRaw,
			     ctx->saHpiSensorThdUpMinorInterpreted,
			     &ctx->saHpiSensorThdUpMinorInterpreted_len,
			     SENSOR_THD_INTER_MAX, NULL, NULL, NULL, 0);
      if (threshold_def) {
      ctx->saHpiSensorThdUpMinorIsReadable =
	((threshold_def->ReadThold & SAHPI_STM_UP_MINOR) ==
	 SAHPI_STM_UP_MINOR) ? MIB_TRUE : MIB_FALSE;

      ctx->saHpiSensorThdUpMinorIsWritable =
	((threshold_def->WriteThold & SAHPI_STM_UP_MINOR) ==
	 SAHPI_STM_UP_MINOR) ? MIB_TRUE : MIB_FALSE;
      ctx->saHpiSensorThdUpMinorIsFixed =
	((threshold_def->FixedThold & SAHPI_STM_UP_MINOR) ==
	 SAHPI_STM_UP_MINOR) ? MIB_TRUE : MIB_FALSE;
       }
      /* END */
      DEBUGMSGTL ((AGENT, "Modify saHpiSensorThdUpMinorTable_ctx: Exit"));
      if (update_entry == MIB_TRUE)
	return AGENT_ENTRY_EXIST;
      return AGENT_NEW_ENTRY;

    }

  DEBUGMSGTL ((AGENT,
	       "Modify saHpiSensorThdUpMinorTable_ctx: Exit (NULL DATA)"));

  return AGENT_ERR_NULL_DATA;
}


int
set_ThdUpMinor (saHpiSensorThdUpMinorTable_context * ctx)
{

  SaHpiSensorThresholdsT thd;
  SaHpiSessionIdT session_id;
  SaErrorT rc;

  DEBUGMSGTL ((AGENT, "set_ThdUpMinor: Entry.\n"));
  if (ctx)
    {

      memset (&thd, 0x00, sizeof (SaHpiSensorThresholdsT));

      rc = getSaHpiSession (&session_id);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		       rc));
	  return rc;
	}
      /*
       * Get the current threshold information
       */
      DEBUGMSGTL ((AGENT, "resource_id: %d, sensor_id: %d\n",
		   ctx->resource_id, ctx->sensor_id));
      rc =
	saHpiSensorThresholdsGet (session_id, ctx->resource_id,
				  ctx->sensor_id, &thd);

      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiSensorThresholdGet fails with return code: %s.\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to saHpiSensorThresholdsGet fails with return code: %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}

      /* Update the correct entry.     */

      if (thd.UpMinor.ValuesPresent & SAHPI_SRF_INTERPRETED)
	{
	  thd.UpMinor.Interpreted.Type = SAHPI_SENSOR_INTERPRETED_TYPE_BUFFER;
	  memcpy (&thd.UpMinor.Interpreted.Value.SensorBuffer,
		  &ctx->saHpiSensorThdUpMinorInterpreted,
		  ctx->saHpiSensorThdUpMinorInterpreted_len);

	}
      if (thd.UpMinor.ValuesPresent & SAHPI_SRF_RAW)
	{
	  thd.UpMinor.Raw = ctx->saHpiSensorThdUpMinorRaw;
	}

      /*
       * Set the thresholds 
       */
      rc = saHpiSensorThresholdsSet (session_id,
				     ctx->resource_id, ctx->sensor_id, &thd);

      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiSensorThresholdSet fails with return code: %s.\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to saHpiSensorThresholdsSet fails with return code: %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}

      /* 
       * Re-read the data. Might be different, so we will need
       * to populate the ctx.
       */

      memset (&thd, 0x00, sizeof (SaHpiSensorThresholdsT));
      rc = saHpiSensorThresholdsGet (session_id,
				     ctx->resource_id, ctx->sensor_id, &thd);

      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to  SensorThresholdGet fails with return code: %s.\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to  SensorThresholdGet fails with return code: %s.\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}
      build_reading_strings (&thd.UpMinor,
			     0,
			     &ctx->saHpiSensorThdUpMinorValuesPresent,
			     &ctx->saHpiSensorThdUpMinorRaw,
			     ctx->saHpiSensorThdUpMinorInterpreted,
			     &ctx->saHpiSensorThdUpMinorInterpreted_len,
			     SENSOR_THD_INTER_MAX, NULL, NULL, NULL, 0);


      DEBUGMSGTL ((AGENT, "set_ThdUpMinor: Exit.\n"));
      return AGENT_ERR_NOERROR;
    }
  DEBUGMSGTL ((AGENT, "set_sensor: Exit.\n"));

  return AGENT_ERR_NULL_DATA;


}

/************************************************************
 * the *_row_copy routine
 */
static int
  saHpiSensorThdUpMinorTable_row_copy
  (saHpiSensorThdUpMinorTable_context * dst,
   saHpiSensorThdUpMinorTable_context * src)
{
  if (!dst || !src)
    return 1;

  /*
   * copy index, if provided
   */
  if (dst->index.oids)
    free (dst->index.oids);
  if (snmp_clone_mem ((void *) &dst->index.oids, src->index.oids,
		      src->index.len * sizeof (oid)))
    {
      dst->index.oids = NULL;
      return 1;
    }
  dst->index.len = src->index.len;


  /*
   * copy components into the context structure
   */
  dst->saHpiSensorThdUpMinorIsReadable = src->saHpiSensorThdUpMinorIsReadable;

  dst->saHpiSensorThdUpMinorIsWritable = src->saHpiSensorThdUpMinorIsWritable;

  dst->saHpiSensorThdUpMinorIsFixed = src->saHpiSensorThdUpMinorIsFixed;

  dst->saHpiSensorThdUpMinorValuesPresent =
    src->saHpiSensorThdUpMinorValuesPresent;

  dst->saHpiSensorThdUpMinorRaw = src->saHpiSensorThdUpMinorRaw;

  memcpy (dst->saHpiSensorThdUpMinorInterpreted,
	  src->saHpiSensorThdUpMinorInterpreted,
	  src->saHpiSensorThdUpMinorInterpreted_len);
  dst->saHpiSensorThdUpMinorInterpreted_len =
    src->saHpiSensorThdUpMinorInterpreted_len;

  dst->resource_id = src->resource_id;
  dst->domain_id = src->domain_id;
  dst->sensor_id = src->sensor_id;
  dst->hash = src->hash;

  return 0;
}



/**
 * the *_extract_index routine
 *
 * This routine is called when a set request is received for an index
 * that was not found in the table container. Here, we parse the oid
 * in the the individual index components and copy those indexes to the
 * context. Then we make sure the indexes for the new row are valid.
 */
int
  saHpiSensorThdUpMinorTable_extract_index
  (saHpiSensorThdUpMinorTable_context * ctx, netsnmp_index * hdr)
{
  /*
   * temporary local storage for extracting oid index
   *
   * extract index uses varbinds (netsnmp_variable_list) to parse
   * the index OID into the individual components for each index part.
   */
  netsnmp_variable_list var_saHpiDomainID;
  netsnmp_variable_list var_saHpiResourceID;
  netsnmp_variable_list var_saHpiSensorIndex;
  int err;

  /*
   * copy index, if provided
   */
  if (hdr)
    {
      netsnmp_assert (ctx->index.oids == NULL);
      if (snmp_clone_mem ((void *) &ctx->index.oids, hdr->oids,
			  hdr->len * sizeof (oid)))
	{
	  return -1;
	}
      ctx->index.len = hdr->len;
    }

  /*
   * initialize variable that will hold each component of the index.
   * If there are multiple indexes for the table, the variable_lists
   * need to be linked together, in order.
   */
  memset (&var_saHpiDomainID, 0x00, sizeof (var_saHpiDomainID));
  var_saHpiDomainID.type = ASN_UNSIGNED;	/* type hint for parse_oid_indexes */

  var_saHpiDomainID.next_variable = &var_saHpiResourceID;

  memset (&var_saHpiResourceID, 0x00, sizeof (var_saHpiResourceID));
  var_saHpiResourceID.type = ASN_UNSIGNED;	/* type hint for parse_oid_indexes */

  var_saHpiResourceID.next_variable = &var_saHpiSensorIndex;

  memset (&var_saHpiSensorIndex, 0x00, sizeof (var_saHpiSensorIndex));
  var_saHpiSensorIndex.type = ASN_UNSIGNED;	/* type hint for parse_oid_indexes */

  var_saHpiSensorIndex.next_variable = NULL;

  /*
   * parse the oid into the individual index components
   */
  err = parse_oid_indexes (hdr->oids, hdr->len, &var_saHpiDomainID);
  if (err == SNMP_ERR_NOERROR)
    {
      /*
       * copy index components into the context structure
       */
      ctx->domain_id = *var_saHpiDomainID.val.integer;
      ctx->resource_id = *var_saHpiResourceID.val.integer;
      ctx->sensor_id = *var_saHpiSensorIndex.val.integer;
    }

  /*
   * parsing may have allocated memory. free it.
   */
  snmp_reset_var_buffers (&var_saHpiDomainID);

  return err;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int
  saHpiSensorThdUpMinorTable_can_delete
  (saHpiSensorThdUpMinorTable_context * undo_ctx,
   saHpiSensorThdUpMinorTable_context * row_ctx, netsnmp_request_group * rg)
{

  return 1;
}

/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiSensorThdUpMinorTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiSensorThdUpMinorTable_context *
saHpiSensorThdUpMinorTable_create_row (netsnmp_index * hdr)
{
  saHpiSensorThdUpMinorTable_context *ctx =
    SNMP_MALLOC_TYPEDEF (saHpiSensorThdUpMinorTable_context);
  if (!ctx)
    return NULL;

  if (saHpiSensorThdUpMinorTable_extract_index (ctx, hdr))
    {
      free (ctx->index.oids);
      free (ctx);
      return NULL;
    }
  ctx->saHpiSensorThdUpMinorIsReadable = MIB_FALSE;
  ctx->saHpiSensorThdUpMinorIsWritable = MIB_FALSE;
  ctx->saHpiSensorThdUpMinorIsFixed = MIB_FALSE;
  ctx->saHpiSensorThdUpMinorValuesPresent = 0;
  return ctx;
}


/************************************************************
 * the *_duplicate row routine
 */
saHpiSensorThdUpMinorTable_context
  * saHpiSensorThdUpMinorTable_duplicate_row
  (saHpiSensorThdUpMinorTable_context * row_ctx)
{
  saHpiSensorThdUpMinorTable_context *dup;

  if (!row_ctx)
    return NULL;

  dup = SNMP_MALLOC_TYPEDEF (saHpiSensorThdUpMinorTable_context);
  if (!dup)
    return NULL;

  if (saHpiSensorThdUpMinorTable_row_copy (dup, row_ctx))
    {
      free (dup);
      dup = NULL;
    }

  return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index *saHpiSensorThdUpMinorTable_delete_row
  (saHpiSensorThdUpMinorTable_context * ctx)
{
  /*
   * netsnmp_mutex_destroy(ctx->lock); 
   */

  if (ctx->index.oids)
    free (ctx->index.oids);

  /*
   * release header
   */
  free (ctx);

  return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void
saHpiSensorThdUpMinorTable_set_reserve1 (netsnmp_request_group * rg)
{
  saHpiSensorThdUpMinorTable_context *row_ctx =
    (saHpiSensorThdUpMinorTable_context *) rg->existing_row;

  netsnmp_variable_list *var;
  netsnmp_request_group_item *current;
  int rc = 0;

  DEBUGMSGTL ((AGENT, "saHpiSensorThdUpMinorTable_set_reserve1. Entry.\n"));
  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;
      rc = SNMP_ERR_NOERROR;

      switch (current->tri->colnum)
	{

	case COLUMN_SAHPISENSORTHDUPMINORISREADABLE:
	case COLUMN_SAHPISENSORTHDUPMINORISWRITABLE:
	case COLUMN_SAHPISENSORTHDUPMINORISFIXED:
	case COLUMN_SAHPISENSORTHDUPMINORVALUESPRESENT:
	  rc = SNMP_ERR_NOTWRITABLE;
	  break;

	case COLUMN_SAHPISENSORTHDUPMINORRAW:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_UNSIGNED,
					       sizeof (row_ctx->
						       saHpiSensorThdUpMinorRaw));
	  break;

	case COLUMN_SAHPISENSORTHDUPMINORINTERPRETED:
	    /** OCTETSTR = ASN_OCTET_STR */
	  if (var->type != ASN_OCTET_STR)
	    {
	      rc = SNMP_ERR_WRONGTYPE;
	    }
	  break;

	default:
		/** We shouldn't get here */
	  rc = SNMP_ERR_GENERR;
	  snmp_log (LOG_ERR, "unknown column in "
		    "saHpiSensorThdUpMinorTable_set_reserve1\n");
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
      rg->status = SNMP_MAX (rg->status, current->ri->status);
    }
  DEBUGMSGTL ((AGENT,
	       "saHpiSensorThdUpMinorTable_set_reserve1. Exit. (rc:%d) \n",
	       rc));
}

void
saHpiSensorThdUpMinorTable_set_reserve2 (netsnmp_request_group * rg)
{
  saHpiSensorThdUpMinorTable_context *row_ctx =
    (saHpiSensorThdUpMinorTable_context *) rg->existing_row;
  netsnmp_request_group_item *current;
  netsnmp_variable_list *var;
  int rc = 0;

  rg->rg_void = rg->list->ri;

  DEBUGMSGTL ((AGENT, "saHpiSensorThdUpMinorTable_set_reserve2. Entry.\n"));
  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;
      rc = SNMP_ERR_NOERROR;

      switch (current->tri->colnum)
	{
	case COLUMN_SAHPISENSORTHDUPMINORINTERPRETED:
	case COLUMN_SAHPISENSORTHDUPMINORRAW:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  if (row_ctx->saHpiSensorThdUpMinorIsWritable == MIB_FALSE)
	    {
	      rc = SNMP_ERR_NOACCESS;
	    }
	  if (row_ctx->saHpiSensorThdUpMinorIsFixed == MIB_TRUE)
	    {
	      rc = SNMP_ERR_NOACCESS;
	    }

	  break;


	default:
		/** We shouldn't get here */
	  netsnmp_assert (0);  /** why wasn't this caught in reserve1? */
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
    }

  DEBUGMSGTL ((AGENT,
	       "saHpiSensorThdUpMinorTable_set_reserve2. Exit (rc:%d).\n",
	       rc));
}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void
saHpiSensorThdUpMinorTable_set_action (netsnmp_request_group * rg)
{
  netsnmp_variable_list *var;
  saHpiSensorThdUpMinorTable_context *row_ctx =
    (saHpiSensorThdUpMinorTable_context *) rg->existing_row;
  netsnmp_request_group_item *current;

  int rc = 0;

  DEBUGMSGTL ((AGENT, "saHpiSensorThdUpMinorTable_set_action. Entry\n"));
  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;

      switch (current->tri->colnum)
	{

	case COLUMN_SAHPISENSORTHDUPMINORRAW:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  row_ctx->saHpiSensorThdUpMinorRaw = *var->val.integer;
	  if (set_ThdUpMinor (row_ctx) != AGENT_ERR_NOERROR)
	    rc = SNMP_ERR_GENERR;
	  break;

	case COLUMN_SAHPISENSORTHDUPMINORINTERPRETED:
	    /** OCTETSTR = ASN_OCTET_STR */
	  memcpy (row_ctx->saHpiSensorThdUpMinorInterpreted,
		  var->val.string, var->val_len);
	  row_ctx->saHpiSensorThdUpMinorInterpreted_len = var->val_len;
	  if (set_ThdUpMinor (row_ctx) != AGENT_ERR_NOERROR)
	    rc = SNMP_ERR_GENERR;

	  break;

	default:
		/** We shouldn't get here */
	  netsnmp_assert (0);  /** why wasn't this caught in reserve1? */
	}
    }


  if (rc)
    {
      netsnmp_set_mode_request_error (MODE_SET_BEGIN,
				      (netsnmp_request_info *) rg->
				      rg_void, rc);
      return;
    }
  DEBUGMSGTL ((AGENT,
	       "saHpiSensorThdUpMinorTable_set_action. Exit (rc: %d)\n", rc));
}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void
saHpiSensorThdUpMinorTable_set_commit (netsnmp_request_group * rg)
{


}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void
saHpiSensorThdUpMinorTable_set_free (netsnmp_request_group * rg)
{

}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void
saHpiSensorThdUpMinorTable_set_undo (netsnmp_request_group * rg)
{

}




/************************************************************
 *
 * Initialize the saHpiSensorThdUpMinorTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiSensorThdUpMinorTable (void)
{
  netsnmp_table_registration_info *table_info;

  if (my_handler)
    {
      snmp_log (LOG_ERR,
		"initialize_table_saHpiSensorThdUpMinorTable_handler called again\n");
      return;
    }

  memset (&cb, 0x00, sizeof (cb));

    /** create the table structure itself */
  table_info = SNMP_MALLOC_TYPEDEF (netsnmp_table_registration_info);

  /*
   * if your table is read only, it's easiest to change the
   * HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY 
   */
  my_handler =
    netsnmp_create_handler_registration
    ("saHpiSensorThdUpMinorTable",
     netsnmp_table_array_helper_handler,
     saHpiSensorThdUpMinorTable_oid,
     saHpiSensorThdUpMinorTable_oid_len, HANDLER_CAN_RWRITE);

  if (!my_handler || !table_info)
    {
      snmp_log (LOG_ERR, "malloc failed in "
		"initialize_table_saHpiSensorThdUpMinorTable_handler\n");
      return;	/** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */

  /*
   * internal indexes
   */
	/** index: saHpiDomainID */
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);
	/** index: saHpiResourceID */
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);
	/** index: saHpiSensorIndex */
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);

  table_info->min_column = saHpiSensorThdUpMinorTable_COL_MIN;
  table_info->max_column = saHpiSensorThdUpMinorTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
  cb.get_value = saHpiSensorThdUpMinorTable_get_value;
  cb.container =
    netsnmp_container_find ("saHpiSensorThdUpMinorTable_primary:"
			    "saHpiSensorThdUpMinorTable:" "table_container");


  cb.create_row = (UserRowMethod *) saHpiSensorThdUpMinorTable_create_row;

  cb.duplicate_row =
    (UserRowMethod *) saHpiSensorThdUpMinorTable_duplicate_row;
  cb.delete_row = (UserRowMethod *) saHpiSensorThdUpMinorTable_delete_row;
  cb.row_copy =
    (Netsnmp_User_Row_Operation *) saHpiSensorThdUpMinorTable_row_copy;

  cb.can_delete =
    (Netsnmp_User_Row_Action *) saHpiSensorThdUpMinorTable_can_delete;

  cb.set_reserve1 = saHpiSensorThdUpMinorTable_set_reserve1;
  cb.set_reserve2 = saHpiSensorThdUpMinorTable_set_reserve2;
  cb.set_action = saHpiSensorThdUpMinorTable_set_action;
  cb.set_commit = saHpiSensorThdUpMinorTable_set_commit;
  cb.set_free = saHpiSensorThdUpMinorTable_set_free;
  cb.set_undo = saHpiSensorThdUpMinorTable_set_undo;

  DEBUGMSGTL (("initialize_table_saHpiSensorThdUpMinorTable",
	       "Registering table saHpiSensorThdUpMinorTable "
	       "as a table array\n"));
  netsnmp_table_container_register (my_handler, table_info, &cb,
				    cb.container, 1);
}

/************************************************************
 * saHpiSensorThdUpMinorTable_get_value
 *
 * This routine is called for get requests to copy the data
 * from the context to the varbind for the request. If the
 * context has been properly maintained, you don't need to
 * change in code in this fuction.
 */
int
saHpiSensorThdUpMinorTable_get_value (netsnmp_request_info * request,
				      netsnmp_index * item,
				      netsnmp_table_request_info * table_info)
{
  netsnmp_variable_list *var = request->requestvb;
  saHpiSensorThdUpMinorTable_context *context =
    (saHpiSensorThdUpMinorTable_context *) item;

#ifdef GET_ROUTINE_CALLS_SNMP_GET
  SaHpiSensorThresholdsT sensor_threshold;
  SaHpiSessionIdT session_id;
  int rc = AGENT_ERR_NOERROR;
  rc = getSaHpiSession (&session_id);
  if (rc != AGENT_ERR_NOERROR)
    {
      DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		   rc));
    }
  rc = saHpiSensorThresholdsGet (session_id,
				 context->resource_id,
				 context->sensor_id, &sensor_threshold);

  if (rc != SA_OK)
    {
      snmp_log (LOG_ERR,
		"Call to saHpiSensorThresholdsGet fails with return code: %s.\n",
		get_error_string (rc));
      DEBUGMSGTL ((AGENT,
		   "Call to  SensorThresholdGet fails with return code: %s.\n",
		   get_error_string (rc)));
      return AGENT_ERR_OPERATION;
    }
  if (rc == AGENT_ERR_NOERROR)
    {
      modify_saHpiSensorThdUpMinorTable_row (context->domain_id,
						 context->resource_id,
						 context->sensor_id,
						 NULL,
						 &sensor_threshold.
						 UpMinor, context);
    }
#endif
  switch (table_info->colnum)
    {

    case COLUMN_SAHPISENSORTHDUPMINORISREADABLE:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorThdUpMinorIsReadable,
				sizeof (context->
					saHpiSensorThdUpMinorIsReadable));
      break;

    case COLUMN_SAHPISENSORTHDUPMINORISWRITABLE:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorThdUpMinorIsWritable,
				sizeof (context->
					saHpiSensorThdUpMinorIsWritable));
      break;

    case COLUMN_SAHPISENSORTHDUPMINORISFIXED:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorThdUpMinorIsFixed,
				sizeof (context->
					saHpiSensorThdUpMinorIsFixed));
      break;

    case COLUMN_SAHPISENSORTHDUPMINORVALUESPRESENT:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiSensorThdUpMinorValuesPresent,
				sizeof (context->
					saHpiSensorThdUpMinorValuesPresent));
      break;

    case COLUMN_SAHPISENSORTHDUPMINORRAW:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      if (context->saHpiSensorThdUpMinorIsReadable == MIB_TRUE)
	snmp_set_var_typed_value (var, ASN_UNSIGNED,
				  (char *) &context->
				  saHpiSensorThdUpMinorRaw,
				  sizeof (context->saHpiSensorThdUpMinorRaw));
      break;

    case COLUMN_SAHPISENSORTHDUPMINORINTERPRETED:
	    /** OCTETSTR = ASN_OCTET_STR */
      if (context->saHpiSensorThdUpMinorIsReadable == MIB_TRUE)
	snmp_set_var_typed_value (var, ASN_OCTET_STR,
				  (char *) &context->
				  saHpiSensorThdUpMinorInterpreted,
				  context->
				  saHpiSensorThdUpMinorInterpreted_len);
      break;

    default:
	    /** We shouldn't get here */
      snmp_log (LOG_ERR, "unknown column in "
		"saHpiSensorThdUpMinorTable_get_value\n");
      return SNMP_ERR_GENERR;
    }
  return SNMP_ERR_NOERROR;
}
