/*
 * (C) Copyright IBM Corp. 2003
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *   Konrad Rzeszutek <konradr@us.ibm.com>
 *
 * Note: this file originally auto-generated by mib2c using
 *       : mib2c.array-user.conf,v 5.15.2.1 2003/02/27 05:59:41 rstory Exp $
 *
 * $Id: saHpiWatchdogTable.c,v 1.12 2004/02/23 21:18:03 konradrzeszutek Exp $
 *
 *
 */
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <net-snmp/library/snmp_assert.h>

#include <saHpiWatchdogTable.h>
#include <SaHpi.h>
#include <hpiSubagent.h>

static netsnmp_handler_registration *my_handler = NULL;
static netsnmp_table_array_callbacks cb;

static oid saHpiWatchdogTable_oid[] = { saHpiWatchdogTable_TABLE_OID };
static size_t saHpiWatchdogTable_oid_len =
OID_LENGTH (saHpiWatchdogTable_oid);

//{ 1, 3, 6, 1, 3, 90, 3, 9, 0 }; 
static oid saHpiWatchdogCount_oid[] = { hpiResources_OID, 10, 0 };
static u_long watchdog_count;

static int
saHpiWatchdogTable_modify_context (SaHpiEntryIdT rdr_id,
				   SaHpiWatchdogRecT * entry,
				   SaHpiRptEntryT * rpt_entry,
				   SaHpiWatchdogT * wdog,
				   oid *, size_t,
				   saHpiWatchdogTable_context * ctx);




int
populate_watchdog (SaHpiEntryIdT rdr_id,
		   SaHpiWatchdogRecT * watchdog,
		   SaHpiRptEntryT * rpt_entry,
		   oid * rdr_entry_oid, size_t rdr_entry_oid_len,
		   oid * watchdog_oid, size_t * watchdog_oid_len)
{

  SaHpiWatchdogT wdog;
  SaHpiSessionIdT session_id;
  int rc = AGENT_ERR_NOERROR;
  oid index_oid[WATCHDOG_INDEX_NR];
  oid column[2];
  netsnmp_index watchdog_index;
#ifdef BUG_873961
  int i = 0;
  netsnmp_void_array *array;
  saHpiWatchdogTable_context *ctx;
#endif
  saHpiWatchdogTable_context *watchdog_context;

  DEBUGMSGTL ((AGENT, "\n\t--- populate_watchdog: Entry.\n"));

  if (watchdog)
    {

      rc = AGENT_ERR_NOERROR;
      watchdog_index.len = WATCHDOG_INDEX_NR;

      // Look at the MIB to find out what the indexs are
      index_oid[0] = rpt_entry->DomainId;
      index_oid[1] = rpt_entry->ResourceId;
      index_oid[2] = watchdog->WatchdogNum;
      // Possible more indexes?

      watchdog_index.oids = (oid *) & index_oid;
      // We are re-populating. Check for existing entries
      watchdog_context = NULL;
      watchdog_context = CONTAINER_FIND (cb.container, &watchdog_index);
#ifdef BUG_873961
      if (!watchdog_context) 
         {
	 // Bug # 873961. We use the 'rdr_id' to check to see if
	 // it is the context. To do so, we have to search for it first.
	 watchdog_index.len = 1;
	 //       // re-using the index_oid.
	 array = CONTAINER_GET_SUBSET (cb.container, &watchdog_index);
		if (array != NULL) 
		{
			if (array->size > 0) 
			{
				for (i = 0; i < array->size; i++) 
				{
					ctx = array->array[i];
					if (ctx->rdr_id == rdr_id) 
					{
						// Found the duplicate entry!
						watchdog_context = ctx;
						index_oid[1] = ctx->resource_id;
						index_oid[2] = ctx->saHpiWatchdogNum;
						DEBUGMSGTL((AGENT,
									"duplicate watchdog entry %d, %d, %d [rdr: %d] found.\n",
									rpt_entry->DomainId,
									rpt_entry->ResourceId,
									watchdog->WatchdogNum,
									rdr_id));
						break;
					}
				}
			}
			free(array->array);
			free(array); array = NULL;
		}
		// restoree it to its previous glory.
		watchdog_index.len =  WATCHDOG_INDEX_NR;
	}
#endif
      // If we don't find it - we create it.
      if (!watchdog_context)
	{
	  // New entry. Add it
	  watchdog_context = saHpiWatchdogTable_create_row (&watchdog_index);
	}
      if (!watchdog_context)
	{
	  snmp_log (LOG_ERR, "Not enough memory for a watchdog row!");
	  return AGENT_ERR_INTERNAL_ERROR;
	}
      // Generate our full OID
      column[0] = 1;
      column[1] = COLUMN_SAHPIWATCHDOGNUM;

      build_full_oid (saHpiWatchdogTable_oid, saHpiWatchdogTable_oid_len,
		      column, 2,
		      &watchdog_index,
		      watchdog_oid, MAX_OID_LEN, watchdog_oid_len);


      // By this stage, watchdog_context surely has something in it.
      // '*_modify_context' does a checksum check to see if 
      // the record needs to be altered, and if so populates with
      // information from RDR and the OIDs passed.
      rc = getSaHpiSession (&session_id);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		       rc));
	  return rc;
	}

      DEBUGMSGTL ((AGENT, "Calling saHpiWatchdogTimerGet with %d\n",
		   watchdog->WatchdogNum));

      rc = saHpiWatchdogTimerGet (session_id, rpt_entry->ResourceId,
				  watchdog->WatchdogNum, &wdog);


      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiWatchdogTimerGet failed with return code: %s.\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to saHpiWatchdogTimerGet failed with rc: %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}

      if (saHpiWatchdogTable_modify_context (rdr_id, watchdog, rpt_entry, &wdog,
					     rdr_entry_oid, rdr_entry_oid_len,
					     watchdog_context)
	  == AGENT_NEW_ENTRY)
	{

	  CONTAINER_INSERT (cb.container, watchdog_context);
	  watchdog_count = CONTAINER_SIZE (cb.container);

	}
      rc = AGENT_ERR_NOERROR;
    }
  else
    rc = AGENT_ERR_OPERATION;

  DEBUGMSGTL ((AGENT, "\n\t--- populate_watchdog. Exit\n"));
  return rc;
}


int
delete_watchdog_row (SaHpiDomainIdT domain_id,
		     SaHpiResourceIdT resource_id, SaHpiWatchdogNumT num)
{

  saHpiWatchdogTable_context *ctx;
  oid index_oid[WATCHDOG_INDEX_NR];
  netsnmp_index index;
  int rc = AGENT_ERR_NOT_FOUND;

  DEBUGMSGTL ((AGENT, "delete_watchdog_row (%d, %d, %d). Entry.\n",
	       domain_id, resource_id, num));

  // Look at the MIB to find out what the indexs are
  index_oid[0] = domain_id;
  index_oid[1] = resource_id;
  index_oid[2] = num;
  // Possible more indexes?
  index.oids = (oid *) & index_oid;
  index.len = WATCHDOG_INDEX_NR;
  ctx = CONTAINER_FIND (cb.container, &index);

  if (ctx)
    {
      CONTAINER_REMOVE (cb.container, ctx);
      saHpiWatchdogTable_delete_row (ctx);
      watchdog_count = CONTAINER_SIZE (cb.container);
      rc = AGENT_ERR_NOERROR;
    }
  DEBUGMSGTL ((AGENT, "delete_watchdog_row. Exit (rc: %d).\n", rc));
  return rc;
}

int
saHpiWatchdogTable_modify_context (SaHpiEntryIdT rdr_id,
				   SaHpiWatchdogRecT * entry,
				   SaHpiRptEntryT * rpt_entry,
				   SaHpiWatchdogT * wdog,
				   oid * rdr_entry, size_t rdr_entry_oid_len,
				   saHpiWatchdogTable_context * ctx)
{
  unsigned int update_entry = MIB_FALSE;
  long hash;

  // Make sure they are valid.
  if (entry && ctx)
    {

      // We are subtracting SaHpiTextBufferT b/c the underlaying HPI
      // library is not zeroing out the memory for not used entries -
      // thus garbage in SaHpiTextBufferT exist,
      hash = calculate_hash_value (entry, sizeof (SaHpiWatchdogRecT));

      DEBUGMSGTL ((AGENT, " Hash value: %d, in ctx: %d\n", hash, ctx->hash));

      if (ctx->hash != 0)
	{
	  // Only do the check if the hash value is something else than zero.
	  // 'zero' value is only for newly created records, and in some
	  // rare instances when the hash has rolled to zero - in which
	  // case we will just consider the worst-case scenario and update
	  // the record and not trust the hash value.
	  if (hash == ctx->hash)
	    {
	      // The same data. No need to change.
	      return AGENT_ENTRY_EXIST;
	    }
	  if (((ctx->resource_id == rpt_entry->ResourceId) &&
	      (ctx->domain_id == rpt_entry->DomainId) &&
	      (ctx->saHpiWatchdogNum == entry->WatchdogNum)) || 
	      (ctx->rdr_id == rdr_id)) {
		  update_entry = MIB_TRUE;
		  DEBUGMSGTL((AGENT,"Updating HotSwap entry. [%d, %d, %d]\n",
					  rpt_entry->DomainId,
					  rpt_entry->ResourceId,
					  entry->WatchdogNum));
	     }
	}

      if (hash == 0)
	hash = 1;
      ctx->hash = hash;

      ctx->saHpiWatchdogRDR_len = rdr_entry_oid_len * sizeof (oid);
      memcpy (ctx->saHpiWatchdogRDR, rdr_entry, ctx->saHpiWatchdogRDR_len);
      ctx->resource_id = rpt_entry->ResourceId;
      ctx->domain_id = rpt_entry->DomainId;
      ctx->rdr_id = rdr_id;
      ctx->saHpiWatchdogNum = entry->WatchdogNum;
      ctx->saHpiWatchdogOem = entry->Oem;
      ctx->saHpiWatchdogTimerReset = MIB_FALSE;
      if (wdog)
	{
	  ctx->saHpiWatchdogLog =
	    (wdog->Log == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;
	  ctx->saHpiWatchdogRunning =
	    (wdog->Running == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;
	  ctx->saHpiWatchdogTimerUse = wdog->TimerUse + 1;
	  ctx->saHpiWatchdogTimerAction = wdog->TimerAction + 1;
	  ctx->saHpiWatchdogPretimerInterrupt = wdog->PretimerInterrupt + 1;
	  ctx->saHpiWatchdogPreTimeoutInterval = wdog->PreTimeoutInterval;
	  ctx->saHpiWatchdogTimerUseExpFlags = wdog->TimerUseExpFlags;
	  ctx->saHpiWatchdogTimerInitialCount = wdog->InitialCount;
	  ctx->saHpiWatchdogTimerPresentCount = wdog->PresentCount;
	}
      if (update_entry == MIB_TRUE)
	      return AGENT_ENTRY_EXIST;
      return AGENT_NEW_ENTRY;
    }

  return AGENT_ERR_NULL_DATA;
}

int
set_timer_reset (saHpiWatchdogTable_context * ctx)
{
  SaHpiWatchdogT wdog;
  SaHpiSessionIdT session_id;
  SaErrorT rc;

  if (ctx)
    {
      if (ctx->saHpiWatchdogRunning == MIB_TRUE)
	{

	  // Get the seesion_id
	  rc = getSaHpiSession (&session_id);
	  if (rc != AGENT_ERR_NOERROR)
	    {
	      DEBUGMSGTL ((AGENT,
			   "Call to getSaHpiSession failed with rc: %d\n",
			   rc));
	      return rc;
	    }

	  DEBUGMSGTL ((AGENT, "Calling saHpiWatchdogTimerReset with %d\n",
		       ctx->saHpiWatchdogNum));

	  rc = saHpiWatchdogTimerReset (session_id,
					ctx->resource_id,
					ctx->saHpiWatchdogNum);

	  if (rc != SA_OK)
	    {
	      snmp_log (LOG_ERR,
			"Call to saHpiWatchdogTimerReset failed with rc: %s\n",
			get_error_string (rc));
	      DEBUGMSGTL ((AGENT,
			   "Call to saHpiWatchdogTimerReset failed with rc: %s\n",
			   get_error_string (rc)));
	      return AGENT_ERR_OPERATION;
	    }
	  // Get the new changes.
	  DEBUGMSGTL ((AGENT, "Calling saHpiWatchdogTimerGet with %d\n",
		       ctx->saHpiWatchdogNum));
	  memset (&wdog, 0x00, sizeof (SaHpiWatchdogT));

	  rc = saHpiWatchdogTimerGet (session_id, ctx->resource_id,
				      ctx->saHpiWatchdogNum, &wdog);
	  if (rc != SA_OK)
	    {
	      snmp_log (LOG_ERR,
			"Call to saHpiWatchdogTimerGet failed with rc: %s.\n",
			get_error_string (rc));
	      DEBUGMSGTL ((AGENT,
			   "Call to saHpiWatchdogTimerGet failed with rc: %s.\n",
			   get_error_string (rc)));
	      return AGENT_ERR_OPERATION;
	    }
	  DEBUGMSGTL ((AGENT, "log: %d, Run: %d\n", wdog.Log, wdog.Running));
	  ctx->saHpiWatchdogLog =
	    (wdog.Log == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;
	  ctx->saHpiWatchdogRunning =
	    (wdog.Running == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;
	  ctx->saHpiWatchdogTimerUse = wdog.TimerUse + 1;
	  ctx->saHpiWatchdogTimerAction = wdog.TimerAction + 1;
	  ctx->saHpiWatchdogPretimerInterrupt = wdog.PretimerInterrupt + 1;
	  ctx->saHpiWatchdogPreTimeoutInterval = wdog.PreTimeoutInterval;
	  ctx->saHpiWatchdogTimerUseExpFlags = wdog.TimerUseExpFlags;
	  ctx->saHpiWatchdogTimerInitialCount = wdog.InitialCount;
	  ctx->saHpiWatchdogTimerPresentCount = wdog.PresentCount;
	  // MIB change. 
	  // IBM-KR: Should this be set to something else, perhaps?
	  ctx->saHpiWatchdogTimerReset = MIB_FALSE;
	  return AGENT_ERR_NOERROR;
	}
      return AGENT_ERR_OPERATION;
    }
  return AGENT_ERR_NULL_DATA;

}

int
set_watchdog (saHpiWatchdogTable_context * ctx)
{

  SaHpiWatchdogT wdog;
  SaHpiSessionIdT session_id;
  SaErrorT rc;

  if (ctx)
    {
      wdog.Log =
	(ctx->saHpiWatchdogLog == MIB_TRUE) ? SAHPI_TRUE : SAHPI_FALSE;
      wdog.Running =
	(ctx->saHpiWatchdogRunning == MIB_TRUE) ? SAHPI_TRUE : SAHPI_FALSE;
      wdog.TimerUse = ctx->saHpiWatchdogTimerUse + 1;
      wdog.TimerAction = ctx->saHpiWatchdogTimerAction + 1;
      wdog.PretimerInterrupt = ctx->saHpiWatchdogPretimerInterrupt + 1;
      wdog.PreTimeoutInterval = ctx->saHpiWatchdogPreTimeoutInterval;
      wdog.TimerUseExpFlags = ctx->saHpiWatchdogTimerUseExpFlags;
      wdog.InitialCount = ctx->saHpiWatchdogTimerInitialCount;
      wdog.PresentCount = ctx->saHpiWatchdogTimerPresentCount;

      // Get the seesion_id
      rc = getSaHpiSession (&session_id);
      if (rc != AGENT_ERR_NOERROR)
	{
	  DEBUGMSGTL ((AGENT, "Call to getSaHpiSession failed with rc: %d\n",
		       rc));
	  return rc;
	}
      DEBUGMSGTL ((AGENT, "Calling saHpiWatchdogTimerSet with %d\n",
		   ctx->saHpiWatchdogNum));
      rc =
	saHpiWatchdogTimerSet (session_id, ctx->resource_id,
			       ctx->saHpiWatchdogNum, &wdog);


      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiWatchdogTimerSet failed with return code: %s\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call saHpiWatchdogTimerSet failed with return code: %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}

      DEBUGMSGTL ((AGENT, "Calling saHpiWatchdogTimerGet with %d\n",
		   ctx->saHpiWatchdogNum));

      rc = saHpiWatchdogTimerGet (session_id, ctx->resource_id,
				  ctx->saHpiWatchdogNum, &wdog);


      if (rc != SA_OK)
	{
	  snmp_log (LOG_ERR,
		    "Call to saHpiWatchdogTimerGet failed with rc: %s\n",
		    get_error_string (rc));
	  DEBUGMSGTL ((AGENT,
		       "Call to saHpiWatchdogTimerGet failed with rc: %s\n",
		       get_error_string (rc)));
	  return AGENT_ERR_OPERATION;
	}

      ctx->saHpiWatchdogLog = (wdog.Log == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;
      ctx->saHpiWatchdogRunning =
	(wdog.Running == SAHPI_TRUE) ? MIB_TRUE : MIB_FALSE;
      ctx->saHpiWatchdogTimerUse = wdog.TimerUse + 1;
      ctx->saHpiWatchdogTimerAction = wdog.TimerAction + 1;
      ctx->saHpiWatchdogPretimerInterrupt = wdog.PretimerInterrupt + 1;
      ctx->saHpiWatchdogPreTimeoutInterval = wdog.PreTimeoutInterval;
      ctx->saHpiWatchdogTimerUseExpFlags = wdog.TimerUseExpFlags;
      ctx->saHpiWatchdogTimerInitialCount = wdog.InitialCount;
      ctx->saHpiWatchdogTimerPresentCount = wdog.PresentCount;
      return AGENT_ERR_NOERROR;
    }
  return AGENT_ERR_NULL_DATA;
}

int
update_watchdog_row (SaHpiDomainIdT domain_id,
		     SaHpiResourceIdT resource_id,
		     SaHpiWatchdogNumT num, SaHpiWatchdogEventT * wdog)
{
  saHpiWatchdogTable_context *ctx;
  oid index_oid[WATCHDOG_INDEX_NR];
  netsnmp_index index;

  // Look at the MIB to find out what the indexs are
  index_oid[0] = domain_id;
  index_oid[1] = resource_id;
  index_oid[2] = num;
  // Possible more indexes?
  index.oids = (oid *) & index_oid;
  index.len = WATCHDOG_INDEX_NR;
  ctx = CONTAINER_FIND (cb.container, &index);

  if (ctx)
    {
      ctx->saHpiWatchdogTimerUse = wdog->WatchdogUse + 1;
      ctx->saHpiWatchdogTimerAction = wdog->WatchdogAction + 1;
      ctx->saHpiWatchdogPretimerInterrupt = wdog->WatchdogPreTimerAction + 1;
      return AGENT_ERR_NOERROR;
    }

  return AGENT_ERR_NOT_FOUND;

}

/************************************************************
 * the *_row_copy routine
 */

static int
saHpiWatchdogTable_row_copy (saHpiWatchdogTable_context * dst,
			     saHpiWatchdogTable_context * src)
{
  if (!dst || !src)
    return 1;

  /*
   * copy index, if provided
   */
  if (dst->index.oids)
    free (dst->index.oids);
  if (snmp_clone_mem ((void *) &dst->index.oids, src->index.oids,
		      src->index.len * sizeof (oid)))
    {
      dst->index.oids = NULL;
      return 1;
    }
  dst->index.len = src->index.len;


  dst->saHpiWatchdogNum = src->saHpiWatchdogNum;

  dst->saHpiWatchdogLog = src->saHpiWatchdogLog;

  dst->saHpiWatchdogRunning = src->saHpiWatchdogRunning;

  dst->saHpiWatchdogTimerUse = src->saHpiWatchdogTimerUse;

  dst->saHpiWatchdogTimerAction = src->saHpiWatchdogTimerAction;

  dst->saHpiWatchdogPretimerInterrupt = src->saHpiWatchdogPretimerInterrupt;

  dst->saHpiWatchdogPreTimeoutInterval = src->saHpiWatchdogPreTimeoutInterval;

  dst->saHpiWatchdogTimerUseExpFlags = src->saHpiWatchdogTimerUseExpFlags;

  dst->saHpiWatchdogTimerInitialCount = src->saHpiWatchdogTimerInitialCount;

  dst->saHpiWatchdogTimerPresentCount = src->saHpiWatchdogTimerPresentCount;

  dst->saHpiWatchdogOem = src->saHpiWatchdogOem;

  dst->saHpiWatchdogRDR_len = src->saHpiWatchdogRDR_len;
  memcpy (dst->saHpiWatchdogRDR, src->saHpiWatchdogRDR,
	  src->saHpiWatchdogRDR_len);

  dst->hash = src->hash;
  dst->resource_id = src->resource_id;
  dst->saHpiWatchdogTimerReset = src->saHpiWatchdogTimerReset;
  return 0;
}



/*
 * the *_extract_index routine
 */
int
saHpiWatchdogTable_extract_index (saHpiWatchdogTable_context * ctx,
				  netsnmp_index * hdr)
{
  /*
   * temporary local storage for extracting oid index
   */
  netsnmp_variable_list var_saHpiDomainID;
  netsnmp_variable_list var_saHpiResourceID;
  netsnmp_variable_list var_saHpiWatchdogNum;
  int err;

  /*
   * copy index, if provided
   */
  if (hdr)
    {
      netsnmp_assert (ctx->index.oids == NULL);
      if (snmp_clone_mem ((void *) &ctx->index.oids, hdr->oids,
			  hdr->len * sizeof (oid)))
	{
	  return -1;
	}
      ctx->index.len = hdr->len;
    }

    /**
     * Create variable to hold each component of the index
     */

  memset (&var_saHpiDomainID, 0x00, sizeof (var_saHpiDomainID));
  var_saHpiDomainID.type = ASN_UNSIGNED;
  var_saHpiDomainID.next_variable = &var_saHpiResourceID;

  memset (&var_saHpiResourceID, 0x00, sizeof (var_saHpiResourceID));
  var_saHpiResourceID.type = ASN_UNSIGNED;
  var_saHpiResourceID.next_variable = &var_saHpiWatchdogNum;

  memset (&var_saHpiWatchdogNum, 0x00, sizeof (var_saHpiWatchdogNum));
  var_saHpiWatchdogNum.type = ASN_UNSIGNED;
  var_saHpiWatchdogNum.next_variable = NULL;

  /*
   * parse the oid into the individual components
   */
  err = parse_oid_indexes (hdr->oids, hdr->len, &var_saHpiDomainID);
  if (err == SNMP_ERR_NOERROR)
    {
      /*
       * copy components into the context structure
       */
      ctx->saHpiWatchdogNum = *var_saHpiWatchdogNum.val.integer;

    }

  /*
   * parsing may have allocated memory. free it.
   */
  snmp_reset_var_buffers (&var_saHpiDomainID);

  return err;
}

/************************************************************
 * the *_can_delete routine is called to determine if a row
 * can be deleted.
 *
 * return 1 if the row can be deleted
 * return 0 if the row cannot be deleted
 */
int
saHpiWatchdogTable_can_delete (saHpiWatchdogTable_context * undo_ctx,
			       saHpiWatchdogTable_context * row_ctx,
			       netsnmp_request_group * rg)
{

  return 1;
}


/************************************************************
 * the *_create_row routine is called by the table handler
 * to create a new row for a given index. If you need more
 * information (such as column values) to make a decision
 * on creating rows, you must create an initial row here
 * (to hold the column values), and you can examine the
 * situation in more detail in the *_set_reserve1 or later
 * states of set processing. Simple check for a NULL undo_ctx
 * in those states and do detailed creation checking there.
 *
 * returns a newly allocated saHpiWatchdogTable_context
 *   structure if the specified indexes are not illegal
 * returns NULL for errors or illegal index values.
 */
saHpiWatchdogTable_context *
saHpiWatchdogTable_create_row (netsnmp_index * hdr)
{
  saHpiWatchdogTable_context *ctx =
    SNMP_MALLOC_TYPEDEF (saHpiWatchdogTable_context);

  if (!ctx)
    return NULL;

  if (saHpiWatchdogTable_extract_index (ctx, hdr))
    {
      free (ctx->index.oids);
      free (ctx);
      return NULL;
    }

  ctx->saHpiWatchdogLog = 0;
  ctx->saHpiWatchdogRunning = 0;
  ctx->saHpiWatchdogTimerUse = 0;
  ctx->saHpiWatchdogTimerAction = 0;
  ctx->saHpiWatchdogPretimerInterrupt = 0;
  ctx->saHpiWatchdogPreTimeoutInterval = 0;
  ctx->saHpiWatchdogTimerUseExpFlags = 0;
  ctx->saHpiWatchdogTimerInitialCount = 0;
  ctx->saHpiWatchdogTimerPresentCount = 0;
  ctx->saHpiWatchdogOem = 0;
  ctx->saHpiWatchdogTimerReset = MIB_FALSE;
  ctx->hash = 0;
  return ctx;
}


/************************************************************
 * the *_duplicate row routine
 */
saHpiWatchdogTable_context *
saHpiWatchdogTable_duplicate_row (saHpiWatchdogTable_context * row_ctx)
{
  saHpiWatchdogTable_context *dup;

  if (!row_ctx)
    return NULL;

  dup = SNMP_MALLOC_TYPEDEF (saHpiWatchdogTable_context);
  if (!dup)
    return NULL;

  if (saHpiWatchdogTable_row_copy (dup, row_ctx))
    {
      free (dup);
      dup = NULL;
    }

  return dup;
}

/************************************************************
 * the *_delete_row method is called to delete a row.
 */
netsnmp_index *
saHpiWatchdogTable_delete_row (saHpiWatchdogTable_context * ctx)
{


  if (ctx->index.oids)
    free (ctx->index.oids);

  free (ctx);

  return NULL;
}


/************************************************************
 * RESERVE is used to check the syntax of all the variables
 * provided, that the values being set are sensible and consistent,
 * and to allocate any resources required for performing the SET.
 * After this stage, the expectation is that the set ought to
 * succeed, though this is not guaranteed. (In fact, with the UCD
 * agent, this is done in two passes - RESERVE1, and
 * RESERVE2, to allow for dependancies between variables).
 *
 * BEFORE calling this routine, the agent will call duplicate_row
 * to create a copy of the row (unless this is a new row; i.e.
 * row_created == 1).
 *
 * next state -> SET_RESERVE2 || SET_FREE
 */
void
saHpiWatchdogTable_set_reserve1 (netsnmp_request_group * rg)
{
  saHpiWatchdogTable_context *row_ctx =
    (saHpiWatchdogTable_context *) rg->existing_row;


  netsnmp_variable_list *var;
  netsnmp_request_group_item *current;

  int rc;

  DEBUGMSGTL ((AGENT, "saHpiWatchdogTable_set_reserve1.Entry\n"));
  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;
      rc = SNMP_ERR_NOERROR;

      switch (current->tri->colnum)
	{
	case COLUMN_SAHPIWATCHDOGNUM:
	case COLUMN_SAHPIWATCHDOGOEM:
	case COLUMN_SAHPIWATCHDOGRDR:
	  rc = SNMP_ERR_NOTWRITABLE;
	  break;
	case COLUMN_SAHPIWATCHDOGLOG:
	    /** TruthValue = ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiWatchdogLog));
	  break;

	case COLUMN_SAHPIWATCHDOGRUNNING:
	    /** TruthValue = ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiWatchdogRunning));

	  break;

	case COLUMN_SAHPIWATCHDOGTIMERUSE:
	    /** INTEGER = ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiWatchdogTimerUse));
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERACTION:
	    /** INTEGER = ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiWatchdogTimerAction));
	  break;

	case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
	    /** INTEGER = ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiWatchdogPretimerInterrupt));
	  break;

	case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_UNSIGNED,
					       sizeof (row_ctx->
						       saHpiWatchdogPreTimeoutInterval));
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_UNSIGNED,
					       sizeof (row_ctx->
						       saHpiWatchdogTimerUseExpFlags));
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_UNSIGNED,
					       sizeof (row_ctx->
						       saHpiWatchdogTimerInitialCount));
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_UNSIGNED,
					       sizeof (row_ctx->
						       saHpiWatchdogTimerPresentCount));
	  break;
	case COLUMN_SAHPIWATCHDOGTIMERRESET:
	  /** TruthValue = ASN_INTEGER */
	  rc = netsnmp_check_vb_type_and_size (var, ASN_INTEGER,
					       sizeof (row_ctx->
						       saHpiWatchdogTimerReset));
	  break;

	default:
		/** We shouldn't get here */
	  rc = SNMP_ERR_GENERR;
	  snmp_log (LOG_ERR, "unknown column in "
		    "saHpiWatchdogTable_set_reserve1\n");
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
      rg->status = SNMP_MAX (rg->status, current->ri->status);
    }
  for (current = rg->list; current; current = current->next)
    {

      // The nice thing about this API is that _row_copy() is called
      // for this row - if the API has matched the index with an
      // already existing entry. We check the 'hash' value. If its
      // 0 the API couldn't find the right context.

      if (((saHpiWatchdogTable_context *) rg->existing_row)->hash == 0)
	{

	  netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri,
					  SNMP_ERR_NOSUCHNAME);
	}
    }
  DEBUGMSGTL ((AGENT, "saHpiWatchdogTable_set_reserve1. Exit\n"));
}

void
saHpiWatchdogTable_set_reserve2 (netsnmp_request_group * rg)
{
  netsnmp_request_group_item *current;
  netsnmp_variable_list *var;
  int rc = SNMP_ERR_NOERROR;

  DEBUGMSGTL ((AGENT, "saHpiWatchdogTable_set_reserve2. Entry\n"));
  rg->rg_void = rg->list->ri;

  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;
      rc = SNMP_ERR_NOERROR;

      switch (current->tri->colnum)
	{
	case COLUMN_SAHPIWATCHDOGTIMERRESET:
	case COLUMN_SAHPIWATCHDOGLOG:
	case COLUMN_SAHPIWATCHDOGRUNNING:
	    /** TruthValue = ASN_INTEGER */
	  rc = netsnmp_check_vb_truthvalue (current->ri->requestvb);
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERUSE:
	  // HPI-MIB::saHpiWatchdogTimerUse 
	  if (((*var->val.integer < 1) ||
	       (*var->val.integer > 7)) && (*var->val.integer != 16))
	    {
	      rc = SNMP_ERR_BADVALUE;
	    }
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERACTION:
	  // HPI-MIB::saHpiWatchdogTimerAction
	  if ((*var->val.integer < 1) || (*var->val.integer > 4))
	    {
	      rc = SNMP_ERR_BADVALUE;
	    }
	  break;

	case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
	    /** INTEGER = ASN_INTEGER */
	  // HPI-MIB::saHpiWatchdogPretimerInterrupt
	  if (((*var->val.integer < 1) ||
	       (*var->val.integer > 4)) && (*var->val.integer != 16))
	    {
	      rc = SNMP_ERR_BADVALUE;
	    }
	  break;

	case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
	  // No restriction. 32-bit unsigned can be anything
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
	  if ((*var->val.integer < 0) || (*var->val.integer > 32))
	    {
	      rc = SNMP_ERR_BADVALUE;
	    }
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:

	  break;

	case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:

	  break;


	default:
		/** We shouldn't get here */
	  netsnmp_assert (0);  /** why wasn't this caught in reserve1? */
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);
    }
  DEBUGMSGTL ((AGENT, "saHpiWatchdogTable_set_reserve2. Exit (rc: %d)\n",
	       rc));

}

/************************************************************
 * Assuming that the RESERVE phases were successful, the next
 * stage is indicated by the action value ACTION. This is used
 * to actually implement the set operation. However, this must
 * either be done into temporary (persistent) storage, or the
 * previous value stored similarly, in case any of the subsequent
 * ACTION calls fail.
 *
 * In your case, changes should be made to row_ctx. A copy of
 * the original row is in undo_ctx.
 */
void
saHpiWatchdogTable_set_action (netsnmp_request_group * rg)
{
  netsnmp_variable_list *var;
  saHpiWatchdogTable_context *row_ctx =
    (saHpiWatchdogTable_context *) rg->existing_row;

  netsnmp_request_group_item *current;
  int rc = SNMP_ERR_NOERROR;

  DEBUGMSGTL ((AGENT, "saHpiWatchdogTable_set_action. Entry\n"));

  for (current = rg->list; current; current = current->next)
    {

      var = current->ri->requestvb;

      switch (current->tri->colnum)
	{

	case COLUMN_SAHPIWATCHDOGLOG:
	    /** TruthValue = ASN_INTEGER */
	  row_ctx->saHpiWatchdogLog = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGRUNNING:
	    /** TruthValue = ASN_INTEGER */
	  row_ctx->saHpiWatchdogRunning = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERUSE:
	    /** INTEGER = ASN_INTEGER */
	  row_ctx->saHpiWatchdogTimerUse = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERACTION:
	    /** INTEGER = ASN_INTEGER */
	  row_ctx->saHpiWatchdogTimerAction = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
	    /** INTEGER = ASN_INTEGER */
	  row_ctx->saHpiWatchdogPretimerInterrupt = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  row_ctx->saHpiWatchdogPreTimeoutInterval = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  row_ctx->saHpiWatchdogTimerUseExpFlags = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  row_ctx->saHpiWatchdogTimerInitialCount = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
	    /** UNSIGNED32 = ASN_UNSIGNED */
	  row_ctx->saHpiWatchdogTimerPresentCount = *var->val.integer;
	  break;

	case COLUMN_SAHPIWATCHDOGTIMERRESET:
	    /** TruthValue = ASN_INTEGER */
	  row_ctx->saHpiWatchdogTimerReset = *var->val.integer;

	  // Set it only when set to MIB_TRUE
	  if (row_ctx->saHpiWatchdogTimerReset == MIB_TRUE)
	    {
	      if (set_timer_reset (row_ctx) != AGENT_ERR_NOERROR)
		rc = SNMP_ERR_GENERR;
	    }
	  break;
	default:
		/** We shouldn't get here */
	  netsnmp_assert (0);  /** why wasn't this caught in reserve1? */
	}

      if (current->tri->colnum != COLUMN_SAHPIWATCHDOGTIMERRESET)
	{
	  DEBUGMSGTL ((AGENT,
		       "saHpiWatchdogTable_set_action: Calling 'set_watchdog'\n"));
	  if (set_watchdog (row_ctx) != AGENT_ERR_NOERROR)
	    {
	      rc = SNMP_ERR_GENERR;
	    }
	}

      if (rc)
	netsnmp_set_mode_request_error (MODE_SET_BEGIN, current->ri, rc);


    }

  DEBUGMSGTL ((AGENT, "saHpiWatchdogTable_set_action. Exit\n"));

}

/************************************************************
 * Only once the ACTION phase has completed successfully, can
 * the final COMMIT phase be run. This is used to complete any
 * writes that were done into temporary storage, and then release
 * any allocated resources. Note that all the code in this phase
 * should be "safe" code that cannot possibly fail (cue
 * hysterical laughter). The whole intent of the ACTION/COMMIT
 * division is that all of the fallible code should be done in
 * the ACTION phase, so that it can be backed out if necessary.
 *
 * BEFORE calling this routine, the agent will update the
 * container (inserting a row if row_created == 1, or removing
 * the row if row_deleted == 1).
 *
 * AFTER calling this routine, the agent will delete the
 * undo_info.
 */
void
saHpiWatchdogTable_set_commit (netsnmp_request_group * rg)
{

}

/************************************************************
 * If either of the RESERVE calls fail, the write routines
 * are called again with the FREE action, to release any resources
 * that have been allocated. The agent will then return a failure
 * response to the requesting application.
 *
 * AFTER calling this routine, the agent will delete undo_info.
 */
void
saHpiWatchdogTable_set_free (netsnmp_request_group * rg)
{

}

/************************************************************
 * If the ACTION phase does fail (for example due to an apparently
 * valid, but unacceptable value, or an unforeseen problem), then
 * the list of write routines are called again, with the UNDO
 * action. This requires the routine to reset the value that was
 * changed to its previous value (assuming it was actually changed),
 * and then to release any resources that had been allocated. As
 * with the FREE phase, the agent will then return an indication
 * of the error to the requesting application.
 *
 * BEFORE calling this routine, the agent will update the container
 * (remove any newly inserted row, re-insert any removed row).
 *
 * AFTER calling this routing, the agent will call row_copy
 * to restore the data in existing_row from the date in undo_info.
 * Then undo_info will be deleted (or existing row, if row_created
 * == 1).
 */
void
saHpiWatchdogTable_set_undo (netsnmp_request_group * rg)
{

}




/************************************************************
 *
 * Initialize the saHpiWatchdogTable table by defining its contents and how it's structured
 */
void
initialize_table_saHpiWatchdogTable (void)
{
  netsnmp_table_registration_info *table_info;

  if (my_handler)
    {
      snmp_log (LOG_ERR,
		"initialize_table_saHpiWatchdogTable_handler called again\n");
      return;
    }

  memset (&cb, 0x00, sizeof (cb));

    /** create the table structure itself */
  table_info = SNMP_MALLOC_TYPEDEF (netsnmp_table_registration_info);

  /*
   * if your table is read only, it's easiest to change the
   * HANDLER_CAN_RWRITE definition below to HANDLER_CAN_RONLY 
   */
  my_handler = netsnmp_create_handler_registration ("saHpiWatchdogTable",
						    netsnmp_table_array_helper_handler,
						    saHpiWatchdogTable_oid,
						    saHpiWatchdogTable_oid_len,
						    HANDLER_CAN_RWRITE);

  if (!my_handler || !table_info)
    {
      snmp_log (LOG_ERR, "malloc failed in "
		"initialize_table_saHpiWatchdogTable_handler\n");
      return;	/** mallocs failed */
    }

    /***************************************************
     * Setting up the table's definition
     */


  /*
   * internal indexes
   */
	/** index: saHpiWatchdogNum */
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);
  netsnmp_table_helper_add_index (table_info, ASN_UNSIGNED);

  table_info->min_column = saHpiWatchdogTable_COL_MIN;
  table_info->max_column = saHpiWatchdogTable_COL_MAX;

    /***************************************************
     * registering the table with the master agent
     */
  cb.get_value = saHpiWatchdogTable_get_value;
  cb.container = netsnmp_container_find ("saHpiWatchdogTable_primary:"
					 "saHpiWatchdogTable:"
					 "table_container");


  cb.can_set = 1;

  cb.create_row = (UserRowMethod *) saHpiWatchdogTable_create_row;

  cb.duplicate_row = (UserRowMethod *) saHpiWatchdogTable_duplicate_row;
  cb.delete_row = (UserRowMethod *) saHpiWatchdogTable_delete_row;
  cb.row_copy = (Netsnmp_User_Row_Operation *) saHpiWatchdogTable_row_copy;

  cb.can_delete = (Netsnmp_User_Row_Action *) saHpiWatchdogTable_can_delete;

  cb.set_reserve1 = saHpiWatchdogTable_set_reserve1;
  cb.set_reserve2 = saHpiWatchdogTable_set_reserve2;
  cb.set_action = saHpiWatchdogTable_set_action;
  cb.set_commit = saHpiWatchdogTable_set_commit;
  cb.set_free = saHpiWatchdogTable_set_free;
  cb.set_undo = saHpiWatchdogTable_set_undo;

  DEBUGMSGTL (("initialize_table_saHpiWatchdogTable",
	       "Registering table saHpiWatchdogTable " "as a table array\n"));
  netsnmp_table_container_register (my_handler, table_info, &cb,
				    cb.container, 1);

  netsnmp_register_read_only_counter32_instance ("watchdog_count",
						 saHpiWatchdogCount_oid,
						 OID_LENGTH
						 (saHpiWatchdogCount_oid),
						 &watchdog_count, NULL);


}

/************************************************************
 * saHpiWatchdogTable_get_value
 */
int
saHpiWatchdogTable_get_value (netsnmp_request_info * request,
			      netsnmp_index * item,
			      netsnmp_table_request_info * table_info)
{
  netsnmp_variable_list *var = request->requestvb;
  saHpiWatchdogTable_context *context = (saHpiWatchdogTable_context *) item;

  switch (table_info->colnum)
    {

    case COLUMN_SAHPIWATCHDOGNUM:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->saHpiWatchdogNum,
				sizeof (context->saHpiWatchdogNum));
      break;

    case COLUMN_SAHPIWATCHDOGLOG:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiWatchdogLog,
				sizeof (context->saHpiWatchdogLog));
      break;

    case COLUMN_SAHPIWATCHDOGRUNNING:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiWatchdogRunning,
				sizeof (context->saHpiWatchdogRunning));
      break;

    case COLUMN_SAHPIWATCHDOGTIMERUSE:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->saHpiWatchdogTimerUse,
				sizeof (context->saHpiWatchdogTimerUse));
      break;

    case COLUMN_SAHPIWATCHDOGTIMERACTION:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiWatchdogTimerAction,
				sizeof (context->saHpiWatchdogTimerAction));
      break;

    case COLUMN_SAHPIWATCHDOGPRETIMERINTERRUPT:
	    /** INTEGER = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiWatchdogPretimerInterrupt,
				sizeof (context->
					saHpiWatchdogPretimerInterrupt));
      break;

    case COLUMN_SAHPIWATCHDOGPRETIMEOUTINTERVAL:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->
				saHpiWatchdogPreTimeoutInterval,
				sizeof (context->
					saHpiWatchdogPreTimeoutInterval));
      break;

    case COLUMN_SAHPIWATCHDOGTIMERUSEEXPFLAGS:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->
				saHpiWatchdogTimerUseExpFlags,
				sizeof (context->
					saHpiWatchdogTimerUseExpFlags));
      break;

    case COLUMN_SAHPIWATCHDOGTIMERINITIALCOUNT:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->
				saHpiWatchdogTimerInitialCount,
				sizeof (context->
					saHpiWatchdogTimerInitialCount));
      break;

    case COLUMN_SAHPIWATCHDOGTIMERPRESENTCOUNT:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->
				saHpiWatchdogTimerPresentCount,
				sizeof (context->
					saHpiWatchdogTimerPresentCount));
      break;

    case COLUMN_SAHPIWATCHDOGTIMERRESET:
	    /** TruthValue = ASN_INTEGER */
      snmp_set_var_typed_value (var, ASN_INTEGER,
				(char *) &context->
				saHpiWatchdogTimerReset,
				sizeof (context->saHpiWatchdogTimerReset));
      break;
    case COLUMN_SAHPIWATCHDOGOEM:
	    /** UNSIGNED32 = ASN_UNSIGNED */
      snmp_set_var_typed_value (var, ASN_UNSIGNED,
				(char *) &context->saHpiWatchdogOem,
				sizeof (context->saHpiWatchdogOem));
      break;
    case COLUMN_SAHPIWATCHDOGRDR:
      /** RowPointer = ASN_OBJECT_ID */
      snmp_set_var_typed_value (var, ASN_OBJECT_ID,
				(char *) &context->saHpiWatchdogRDR,
				context->saHpiWatchdogRDR_len);
      break;
    default:
	    /** We shouldn't get here */
      snmp_log (LOG_ERR, "unknown column in "
		"saHpiWatchdogTable_get_value\n");
      return SNMP_ERR_GENERR;
    }
  return SNMP_ERR_NOERROR;
}
